#include "gpu_code.h"
#include <vector_types.h>
#include "mt19937ar.h" //random number generator
#include <cassert>

#define AND_PC_MASK           15L   // 0000 0 1111
#define AND_SLOT_MASK         16L   // 0000 1 0000
#define AND_MYPLACE_MASK     480L   // 1111 0 0000
#define AND_STATE_MASK       511L   // 1111 1 1111

#define AND_OFFSET_SLOT        4L
#define AND_OFFSET_MYPLACE     5L
#define AND_STATE_BITSHIFT     9L

#define  AND_NEXTSHIFT        54L

#define  NCS (pc==0)
#define  P1  (pc==1)
#define  P2  (pc==2)
#define  P3  (pc==3)
#define  CS  (pc==4)

#define  LAMP_PC_MASK           15L   // 0000 0 1111
#define  LAMP_SLOT_MASK         16L   // 0000 1 0000
#define  LAMP_J_MASK           480L   // 1111 0 0000
#define  LAMP_STATE_MASK       511L   // 1111 1 1111

#define  LAMP_OFFSET_SLOT       4L
#define  LAMP_OFFSET_J          5L
#define  LAMP_STATE_BITSHIFT    9L

#define  LAMP_X_MASK            1008806316530991104L  // 0000 1110 | 0000 0000 | 0000 0000 | 0000 0000 | 0000 0000 | 0000 0000 | 0000 0000 | 0000 0000
#define  LAMP_Y_MASK            8070450532247928832L  // 0111 0000 | 0000 0000 | 0000 0000 | 0000 0000 | 0000 0000 | 0000 0000 | 0000 0000 | 0000 0000

#define  LAMP_X_SHIFT           57L
#define  LAMP_Y_SHIFT           60L

#define  LAMP_MAX_Y              7L
#define  LAMP_MAX_X              7L

#define  LAMP_NCS  (pc==0)
#define  LAMP_CS   (pc==1)
#define  LAMP_Q1   (pc==2)
#define  LAMP_Q2   (pc==3)
#define  LAMP_Q22  (pc==4)
#define  LAMP_P    (pc==5)
#define  LAMP_Q3   (pc==6)
#define  LAMP_Q4   (pc==7)
#define  LAMP_Q5   (pc==8)
#define  LAMP_E1   (pc==9)


#define  PET_PC_MASK            7L   //  000 000 000 000 111
#define  PET_POS_MASK          56L   //  000 000 000 111 000
#define  PET_STEP_MASK        448L   //  000 000 111 000 000
#define  PET_J_MASK          3584L   //  000 111 000 000 000
#define  PET_K_MASK         28672L   //  111 000 000 000 000


#define  PET_STATE_MASK     32767L   //  111 111 111 111 111

#define  PET_OFFSET_POS        3L
#define  PET_OFFSET_STEP       6L
#define  PET_OFFSET_J          9L
#define  PET_OFFSET_K         12L

#define  PET_STATE_BITSHIFT   15L


#define  PET_NCS    (pc==0)
#define  PET_CS     (pc==1)
#define  PET_WAIT   (pc==2)
#define  PET_Q2     (pc==3)
#define  PET_Q3     (pc==4)


#define rot(x,k)(((x)<<(k))|((x)>>(32-(k))))
#define mix(a,b,c)\
{\
    a-=c; a^=rot(c,4); c+=b;\
    b-=a; b^=rot(a,6); a+=c;\
    c-=b; c^=rot(b,8); b+=a;\
    a-=c; a^=rot(c,16); c+=b;\
    b-=a; b^=rot(a,19); a+=c;\
    c-=b; c^=rot(b,4); b+=a;\
}

__device__ __constant__ int consCFG[16];
__device__ __constant__ int consOrder[96];
__device__ __constant__ unsigned long long int consRnd[100];


//-----
//Hash function generation/assignment
//-----

void GenerateFunctions(uint2 *constants) {
  bool regenerate = true;

  while (regenerate) {
    regenerate = false;

    // Generate a set of hash function constants for this build attempt.
    // the +1 is for stash constants
    for (unsigned i = 0 ; i < (NUMHASHFUNC+1); ++i) {
      unsigned new_a = genrand_int32() % kPrimeDivisor;
      constants[i].x = (1 > new_a ? 1 : new_a);
      constants[i].y = genrand_int32() % kPrimeDivisor;
    }

/*//Probably a good idea to re-implement, but it likely doesn't need all of these parameters
#ifdef FORCEFULLY_GENERATE_NO_CYCLES
    // Ensure that every key gets N different slots.
    regenerate = CheckAssignedSameSlot(N, num_keys, d_keys, table_size, constants);
#endif*/
  }

}




inline __device__ __host__
unsigned hash_function_inner(const uint2    constants,
                             unsigned long long int key, int table_size) {
#if 1                             
  // Fast version.                             
  return (((constants.x ^ key) + constants.y) % kPrimeDivisor)%table_size;
#else
  // Slow version.
  return (((unsigned unsigned long long)constants.x * key + constants.y) % kPrimeDivisor)%table_size;
#endif
}                             



//! Computes the value of a hash function for a given key.
/*! \param[in] functions        All of the constants used by the hash functions.
  ! \param[in] which_function   Which hash function is being used.
  ! \param[in] key              Key being hashed.
  ! \returns                    The value of a hash function with a given key.
 */
inline __device__ __host__
unsigned hash_function(Functions * functions,
                       int which_function,
                       unsigned long long int key, int table_size) {
  return hash_function_inner(functions->constants[which_function], key, table_size);
}




//-----
//Existing functions for on-the-fly exploration
//-----

__device__ int num_successors(unsigned long long int * vertex_list, unsigned long long int state)
{
           return vertex_list[state+1] - vertex_list[state];
}

__device__ int successor(unsigned long long int * vertex_list, unsigned long long int * edges_list, unsigned long long int state, unsigned long long int isucc)                  
{
            return edges_list[vertex_list[state] + isucc];

}

__device__ unsigned int select_random_worker (curandState *state, int w)
{
     curandState localState = state[w];
     return curand(&localState) % N;
}

__device__ unsigned long long int extract_state_from_queue (Queue * queue, unsigned long long int * queues_slots)
{
     unsigned long long int index, state;
     index                      = (queue->start != queue->stop);
     queue->stop               -= index;
     state                      = queues_slots[queue->stop];
     queues_slots[queue->stop]  = 18446744073709551615;//because unsigned
     return state;
}

__device__ void insert_state_in_queue (Queue * queue, unsigned long long int * queues_slots, unsigned long long int state)
{
      queues_slots[queue->stop] = state;
      queue->stop++;
      queues_slots[queue->stop] = 18446744073709551615;//because unsigned
}

__device__ int modelcheckState (unsigned long long int state)
{
      
      return (state>10000000);
}

__global__ void setup_kernel(curandState *state) {
       
            int id = blockIdx.x * blockDim.x + threadIdx.x;

            /* Each thread gets same seed, a different sequence number , no offset */
            curand_init(1234, id, 0, &state[id]); 
}



//-----
//Hashing device functions
//-----

__device__ bool add(int address, bool * table)
{
    bool u=0;
    u=table[address];  //this is what was hashed at this location (or 0 if nothing was here)

    table[address]=1;  //put our current value in the table

    return (u);//what was in the address already (possibly 0)
}


//new
__device__ bool checkState(int i, unsigned long long int value, Functions * functions, bool * table, int table_size)
{
    int hashed_value=hash_function(functions, i, value, table_size);
    return (table[hashed_value]==value);
}




__device__ int findState(int hashed_value, unsigned long long int value, Functions * functions, unsigned long long int * table, int table_size)
{
    int i=0;
    while(!hashed_value==hash_function(functions, i, value, table_size)){i++;}
    return (i);
}

__device__ volatile int empty;


//-----
// This is our lock-free on the fly (greater than 2) function with hashing added
//-----
/*__global__ void stateSpaceExploration (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				unsigned long long int * table,
				unsigned long long int * stashtable,
				int table_size,
				volatile int * gpu_ArrayIn, 
                                volatile int * gpu_ArrayOut)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
	unsigned long long int       evic_state=1;//must start at 1 because 0 stops the while loop
        //unsigned long long int       devic_state=1;//used as part of an unnecessary check
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
	int chain=0;
	bool visitedState= false;


        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];

 
     
	do 
        {
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
	
			for(iproc=0; iproc<numproc; iproc++)
			{
				
				succ_state=execBit(iproc, numproc, state);//calls the device function that returns a proper successor
				
				i=0;
				visitedState=false;
				visitedState=checkStash(succ_state, function_set->constants[NUMHASHFUNC], stashtable, table_size);
				while(i<NUMHASHFUNC)  //check if we have seen this state before
			    	{
					visitedState=checkState(i, succ_state, function_set, table, table_size);
					if(visitedState){break;}//we have, stop
					i++;
			    	}
			    	if (i==NUMHASHFUNC)//inelegant way to suggest we only do this if we have not already visited this state
			    	{
					hashed_value=hash_function(function_set, 0, succ_state, table_size);//put it in its first feasible slot

					//property check goes here, breaking if violation is found
					evic_state=add(hashed_value, succ_state, table);  //add global state vector to hash table, return evicted value (possibly 0)
					//evic_state=(evic_state == succ_state ? 0 : evic_state);//a nice check, but doesn't do anything useful in practice

					while(evic_state!=0 && chain<MAXITERATIONS)
					{
				  		j=0;//yes, I have co-opted j from the sync, but it should not interfere with that
				  		j=findState(hashed_value, evic_state, function_set, table, table_size);//this is the same hashed_value out of the last add() call
				  		hashed_value=hash_function(function_set, (j+1%NUMHASHFUNC), evic_state, table_size);//this is the incremented hashed_value
				  		evic_state = add(hashed_value,evic_state, table);
				  		chain++;
					}
					if(chain>=MAXITERATIONS)//Ideally this wouldn't be an if, but I digress
					{
				  		hashed_value=stash_hash_function(function_set->constants[NUMHASHFUNC], evic_state, table_size);//need to call stash function
				  		evic_state = stashadd(hashed_value, evic_state, stashtable);
						//evic_state = (devic_state == evic_state ? 0 : devic_state); //a nice check, but doesn't do anything useful in practice
					}
					if(evic_state!=0){break;}//if this happens, we need to tune our hash functions

					//put generated successor in random "mailbox" for next round
					w_new = curand(&localState) % N;
					lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;
					insert_state_in_queue (lwqueue, queues_slots, succ_state);


			   	}	
			}

		     }
		}
	    t = 1 - t;             
            
            gpu_lock_free_sync(10, gpu_ArrayIn, gpu_ArrayOut);
            
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            
            gpu_lock_free_sync(20, gpu_ArrayIn, gpu_ArrayOut);      
     } while (!empty);
}*/












__device__ bool checkRnd(unsigned long long int state)
{
    int match = 0;
    int inc = 0;
    for(inc = 0; inc < 100; inc++)
    {
	match += (consRnd[inc]==state);
    }

    return (match>0);
}


/*__device__ bool checkRnd(unsigned long long int state)
{
    bool match=0;
    int counter=0;

counter+=(2724573682 == state);
counter+=(1492216537 == state);
counter+=(1132903362 == state);
counter+=(3242996771 == state);
counter+=(4275361144 == state);
counter+=(4078033071 == state);
counter+=(651086860 == state);
counter+=(1240057350 == state);
counter+=(2550607503 == state);
counter+=(2269235132 == state);

counter+=(1442100140 == state);
counter+=(4006204051 == state);
counter+=(1126484376 == state);
counter+=(2156529078 == state);
counter+=(4060407080 == state);
counter+=(951507193 == state);
counter+=(3130009906 == state);
counter+=(1912423075 == state);
counter+=(1724117802 == state);
counter+=(2278866624 == state);

counter+=(2256772788 == state);
counter+=(177865216 == state);
counter+=(1886131292 == state);
counter+=(3784743381 == state);
counter+=(3077770975 == state);
counter+=(3960460535 == state);
counter+=(2272212880 == state);
counter+=(3099642104 == state);
counter+=(4028196542 == state);
counter+=(4276981009 == state);

counter+=(3148573739 == state);
counter+=(3139047893 == state);
counter+=(3087141048 == state);
counter+=(4121465502 == state);
counter+=(3392086205 == state);
counter+=(2239781229 == state);
counter+=(1449886365 == state);
counter+=(242469469 == state);
counter+=(611149635 == state);
counter+=(2945944980 == state);

counter+=(2115457847 == state);
counter+=(1545958569 == state);
counter+=(2666702071 == state);
counter+=(3576398588 == state);
counter+=(951268287 == state);
counter+=(907061744 == state);
counter+=(1874498046 == state);
counter+=(3960419408 == state);
counter+=(1417492509 == state);
counter+=(731456838 == state);

counter+=(3450441481 == state);
counter+=(376155839 == state);
counter+=(1642734799 == state);
counter+=(1486906590 == state);
counter+=(22693836 == state);
counter+=(382261787 == state);
counter+=(2295770959 == state);
counter+=(3190043895 == state);
counter+=(1277341516 == state);
counter+=(3926948725 == state);

counter+=(2890544779 == state);
counter+=(209589084 == state);
counter+=(872053432 == state);
counter+=(749410935 == state);
counter+=(3756820866 == state);
counter+=(4032809493 == state);
counter+=(36191491 == state);
counter+=(1195901925 == state);
counter+=(89929818 == state);
counter+=(3836501419 == state);

counter+=(4250701303 == state);
counter+=(2274507151 == state);
counter+=(2334260137 == state);
counter+=(873519419 == state);
counter+=(2075275443 == state);
counter+=(400221074 == state);
counter+=(3837981428 == state);
counter+=(2879127877 == state);
counter+=(9614535 == state);
counter+=(2070503392 == state);

counter+=(1161888038 == state);
counter+=(170155645 == state);
counter+=(1547432817 == state);
counter+=(3144672524 == state);
counter+=(59148843 == state);
counter+=(1943332635 == state);
counter+=(334126155 == state);
counter+=(4119439421 == state);
counter+=(3304030197 == state);
counter+=(206142579 == state);

counter+=(1963299479 == state);
counter+=(2172004172 == state);
counter+=(1218289947 == state);
counter+=(1376476510 == state);
counter+=(1658628670 == state);
counter+=(1032954288 == state);
counter+=(1327745764 == state);
counter+=(80222616 == state);
counter+=(263530022 == state);
counter+=(3977529182 == state);

match=(counter>0);

    return (match);
}*/















//-----
//  With rep table, AB mix, gen32, all ND random order (inner loop)
//
//  Done in FPGA style
//-----
__global__ void stateSpaceExplorationFPGAPDS (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a, unsigned long long int mix_b)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
        unsigned long long int       mix_state=0;
	unsigned long long int unmix_a=mix_a;
	unsigned long long int unmix_b=mix_b;
	unsigned long long int randND=mix_a;
	unsigned long long int selection=0;
	int sel=0;
	int nd = 0;
	int nd_sel=0;
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
	int rep_value=0;
        //int unmixed=0;
	__shared__ uint8_t s_table[49100];
	uint8_t visitedState= 0;//not sure what all 0s is in char

        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];

	//BUG SOURCE
	//for(i=0;i<49100;i++)
	//{
	//s_table[i] = table[i];
	//}

        nd_sel=randND%20;//setting random ND order

	for (i = threadIdx.x; i < 49100; i += blockDim.x) 
    		s_table[i] = 0;
	__syncthreads();

 
     
	do 
        {
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
			for(iproc=0; iproc<numproc; iproc++){
			//for(nd=0; nd<4; nd++){
				nd=randND%4;
				succ_state=execBitGen32(iproc, nd, state);//calls the device function that returns a proper successor
				

				w_new = curand(&localState) % N;
				lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;

				visitedState=0;//in char
				selection=0;//same
				mix_state=succ_state;
				mix_a=unmix_a;
				mix_b=unmix_b;
				selection=(mix(mix_a, mix_b, mix_state));
				hashed_value=(selection/8)%table_size;
				sel=selection%8;//to select bit
				randND=mix_a;

				if(checkRnd(succ_state))
				{
					rep_value=hashed_value+(50000*sel);
					rep_table[rep_value]=succ_state;  //FOR REPORTING ONLY!  DISABLE ON REAL RUNS
				}
				else
				{
					//if((lwqueue->stop) < ((1-t)*NxNxMAX_QUEUE_NUM_SLOTS+w_new*NxMAX_QUEUE_NUM_SLOTS+(w+1)*MAX_QUEUE_NUM_SLOTS))
					if((lwqueue->stop) < 2*NxNxMAX_QUEUE_NUM_SLOTS)
					{
						//pull copy
						visitedState=s_table[hashed_value];  //this is what was hashed at this location (or 0 if nothing was here)
						s_table[hashed_value]|=(1<<sel);  //put our current value in the table (bitwise set)

						if((visitedState&(1<<sel))==0)//if this bit was not previously visited
						{
							//put generated successor in random "mailbox" for next round
							insert_state_in_queue (lwqueue, queues_slots, succ_state);
						}
					}//slots available
				}//else
			   		
			}//}//nd and iproc

		     }//while
		}//q
	    t = 1 - t;             
            

            __syncthreads();
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            

	    __syncthreads();     
     } while (!empty);
	for(i=0;i<49100;i++)
	{
	table[i] = s_table[i];
	}
}









//-----
//  With rep table, AB mix, gen32, all ND random order (inner loop)
//
//  Done in FPGA style
//-----
__global__ void stateSpaceExplorationFPGAEven (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a, unsigned long long int mix_b)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
        unsigned long long int       mix_state=0;
	unsigned long long int unmix_a=mix_a;
	unsigned long long int unmix_b=mix_b;
	unsigned long long int randND=mix_a;
	unsigned long long int selection=0;
	int sel=0;
	int nd = 0;
	int nd_sel=0;
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
	int rep_value=0;
        //int unmixed=0;
	__shared__ uint8_t s_table[49100];
	uint8_t visitedState= 0;//not sure what all 0s is in char

        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];

	//BUG SOURCE
	//for(i=0;i<49100;i++)
	//{
	//s_table[i] = table[i];
	//}

        nd_sel=randND%20;//setting random ND order

	for (i = threadIdx.x; i < 49100; i += blockDim.x) 
    		s_table[i] = 0;
	__syncthreads();

 
     
	do 
        {
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
			for(iproc=0; iproc<numproc; iproc+=2){
			for(nd=0; nd<4; nd++)
			{
				//nd=randND%4;
				succ_state=execBitGen32(iproc, consOrder[(nd_sel*4)+nd], state);//calls the device function that returns a proper successor
				

				w_new = curand(&localState) % N;
				lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;

				visitedState=0;//in char
				selection=0;//same
				mix_state=succ_state;
				mix_a=unmix_a;
				mix_b=unmix_b;
				selection=(mix(mix_a, mix_b, mix_state));
				hashed_value=(selection/8)%table_size;
				sel=selection%8;//to select bit
				//randND=mix_a;

				if(checkRnd(succ_state))
				{
					rep_value=hashed_value+(50000*sel);
					rep_table[rep_value]=succ_state;  //FOR REPORTING ONLY!  DISABLE ON REAL RUNS
				}
				else
				{
					//if((lwqueue->stop) < ((1-t)*NxNxMAX_QUEUE_NUM_SLOTS+w_new*NxMAX_QUEUE_NUM_SLOTS+(w+1)*MAX_QUEUE_NUM_SLOTS))
					if((lwqueue->stop) < 2*NxNxMAX_QUEUE_NUM_SLOTS)
					{
						//pull copy
						visitedState=s_table[hashed_value];  //this is what was hashed at this location (or 0 if nothing was here)
						s_table[hashed_value]|=(1<<sel);  //put our current value in the table (bitwise set)

						if((visitedState&(1<<sel))==0)//if this bit was not previously visited
						{
							//put generated successor in random "mailbox" for next round
							insert_state_in_queue (lwqueue, queues_slots, succ_state);
						}
					}//slots available
				}//else
			   		
			}}//nd and iproc

		     }
		}
	    t = 1 - t;             
            

            __syncthreads();
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            

	    __syncthreads();     
     } while (!empty);
	for(i=0;i<49100;i++)
	{
	table[i] = s_table[i];
	}
}









//-----
//  With rep table, AB mix, gen32, all ND random order (inner loop)
//
//  Done in FPGA style
//-----
__global__ void stateSpaceExplorationFPGAOdd (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a, unsigned long long int mix_b)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
        unsigned long long int       mix_state=0;
	unsigned long long int unmix_a=mix_a;
	unsigned long long int unmix_b=mix_b;
	unsigned long long int randND=mix_a;
	unsigned long long int selection=0;
	int sel=0;
	int nd = 0;
	int nd_sel=0;
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
	int rep_value=0;
        //int unmixed=0;
	__shared__ uint8_t s_table[49100];
	uint8_t visitedState= 0;//not sure what all 0s is in char

        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];

	//BUG SOURCE
	//for(i=0;i<49100;i++)
	//{
	//s_table[i] = table[i];
	//}

        nd_sel=randND%20;//setting random ND order

	for (i = threadIdx.x; i < 49100; i += blockDim.x) 
    		s_table[i] = 0;
	__syncthreads();

 
     
	do 
        {
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
			for(iproc=1; iproc<numproc; iproc+=2){
			for(nd=0; nd<4; nd++)
			{
				//nd=randND%4;
				succ_state=execBitGen32(iproc, consOrder[(nd_sel*4)+nd], state);//calls the device function that returns a proper successor
				

				w_new = curand(&localState) % N;
				lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;

				visitedState=0;//in char
				selection=0;//same
				mix_state=succ_state;
				mix_a=unmix_a;
				mix_b=unmix_b;
				selection=(mix(mix_a, mix_b, mix_state));
				hashed_value=(selection/8)%table_size;
				sel=selection%8;//to select bit
				//randND=mix_a;

				if(checkRnd(succ_state))
				{
					rep_value=hashed_value+(50000*sel);
					rep_table[rep_value]=succ_state;  //FOR REPORTING ONLY!  DISABLE ON REAL RUNS
				}
				else
				{
					//if((lwqueue->stop) < ((1-t)*NxNxMAX_QUEUE_NUM_SLOTS+w_new*NxMAX_QUEUE_NUM_SLOTS+(w+1)*MAX_QUEUE_NUM_SLOTS))
					if((lwqueue->stop) < 2*NxNxMAX_QUEUE_NUM_SLOTS)
					{
						//pull copy
						visitedState=s_table[hashed_value];  //this is what was hashed at this location (or 0 if nothing was here)
						s_table[hashed_value]|=(1<<sel);  //put our current value in the table (bitwise set)

						if((visitedState&(1<<sel))==0)//if this bit was not previously visited
						{
							//put generated successor in random "mailbox" for next round
							insert_state_in_queue (lwqueue, queues_slots, succ_state);
						}
					}//slots available
				}//else
			   		
			}}//nd and iproc

		     }
		}
	    t = 1 - t;             
            

            __syncthreads();
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            

	    __syncthreads();     
     } while (!empty);
	for(i=0;i<49100;i++)
	{
	table[i] = s_table[i];
	}
}













//-----
//  With rep table, AB mix, gen32, all ND random order (inner loop)
//
//  Done in FPGA style
//-----
__global__ void stateSpaceExplorationFPGA (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a, unsigned long long int mix_b)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
        unsigned long long int       mix_state=0;
	unsigned long long int unmix_a=mix_a;
	unsigned long long int unmix_b=mix_b;
	unsigned long long int randND=mix_a;
	unsigned long long int selection=0;
	int sel=0;
	int nd = 0;
	int nd_sel=0;
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
	int rep_value=0;
        //int unmixed=0;
	__shared__ uint8_t s_table[49100];
	uint8_t visitedState= 0;//not sure what all 0s is in char

        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];

	//BUG SOURCE
	//for(i=0;i<49100;i++)
	//{
	//s_table[i] = table[i];
	//}

        nd_sel=randND%20;//setting random ND order

	for (i = threadIdx.x; i < 49100; i += blockDim.x) 
    		s_table[i] = 0;
	__syncthreads();

 
     
	do 
        {
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
			for(iproc=0; iproc<numproc; iproc++){
			for(nd=0; nd<4; nd++)
			{
				//nd=randND%4;
				//succ_state=execBitGen32(iproc, consOrder[(nd_sel*4)+nd], state);//calls the device function that returns a proper successor
				//succ_state=execBit(iproc, numproc, state);
				//succ_state=execBitAE(iproc, numproc, state);
				succ_state=execBitAO(iproc, numproc, state);
				//succ_state=execBitPeterson1(iproc, (consOrder[(nd_sel*4)+nd]%2), numproc, state);
				//succ_state=execBitPeterson2(iproc, (consOrder[(nd_sel*4)+nd]%2), numproc, state);
				//succ_state=execBitPeterson3(iproc, (consOrder[(nd_sel*4)+nd]%2), numproc, state);

				w_new = curand(&localState) % N;
				lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;

				visitedState=0;//in char
				selection=0;//same
				mix_state=succ_state;
				mix_a=unmix_a;
				mix_b=unmix_b;
				selection=(mix(mix_a, mix_b, mix_state));
				hashed_value=(selection/8)%table_size;
				sel=selection%8;//to select bit
				//randND=mix_a;

//send all
rep_value=hashed_value+(50000*sel);
rep_table[rep_value]=succ_state;
//close send all


				if(checkRnd(succ_state))
				{
					rep_value=hashed_value+(50000*sel);
					rep_table[rep_value]=succ_state;  //FOR REPORTING ONLY!  DISABLE ON REAL RUNS
				}
				else
				{
					//if((lwqueue->stop) < ((1-t)*NxNxMAX_QUEUE_NUM_SLOTS+w_new*NxMAX_QUEUE_NUM_SLOTS+(w+1)*MAX_QUEUE_NUM_SLOTS))
					if((lwqueue->stop) < 2*NxNxMAX_QUEUE_NUM_SLOTS)
					{
						//pull copy
						visitedState=s_table[hashed_value];  //this is what was hashed at this location (or 0 if nothing was here)
						s_table[hashed_value]|=(1<<sel);  //put our current value in the table (bitwise set)

						if((visitedState&(1<<sel))==0)//if this bit was not previously visited
						{
							//put generated successor in random "mailbox" for next round
							insert_state_in_queue (lwqueue, queues_slots, succ_state);
						}
					}//slots available
				}//else
			   		
			}}//nd and iproc

		     }
		}
	    t = 1 - t;             
            

            __syncthreads();
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            

	    __syncthreads();     
     } while (!empty);
	for(i=0;i<49100;i++)
	{
	table[i] = s_table[i];
	}
}
















//-----
//  With rep table, AB mix, gen32
//-----
__global__ void stateSpaceExplorationAB32BitOdd (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a, unsigned long long int mix_b)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
        unsigned long long int       mix_state=0;
	unsigned long long int unmix_a=mix_a;
	unsigned long long int unmix_b=mix_b;
	unsigned long long int randND=mix_a;
	unsigned long long int selection=0;
	int sel=0;
	int nd = 0;
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
	int rep_value=0;
        //int unmixed=0;
	__shared__ uint8_t s_table[49100];
	uint8_t visitedState= 0;//not sure what all 0s is in char

        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];

	//BUG SOURCE
	//for(i=0;i<49100;i++)
	//{
	//s_table[i] = table[i];
	//}

	for (i = threadIdx.x; i < 49100; i += blockDim.x) 
    		s_table[i] = 0;
	__syncthreads();

 
     
	do 
        {
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
	
			for(iproc=1; iproc<numproc; iproc+=2)
			{
				nd=randND%4;
				succ_state=execBitGen32(iproc, nd, state);//calls the device function that returns a proper successor
				
				visitedState=0;//in char
				selection=0;//same

				mix_state=succ_state;
				mix_a=unmix_a;
				mix_b=unmix_b;
				selection=(mix(mix_a, mix_b, mix_state));
				hashed_value=(selection/8)%table_size;
				sel=selection%8;//to select bit
				randND=mix_a;

				//pull copy
				visitedState=s_table[hashed_value];  //this is what was hashed at this location (or 0 if nothing was here)
				
				s_table[hashed_value]|=(1<<sel);  //put our current value in the table (bitwise set)

				if((visitedState&(1<<sel))==0)//if this bit was not previously visited
				{
					//put generated successor in random "mailbox" for next round
					rep_value=hashed_value+(50000*sel);
					rep_table[rep_value]=succ_state;  //FOR REPORTING ONLY!  DISABLE ON REAL RUNS
					w_new = curand(&localState) % N;
					lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;
					insert_state_in_queue (lwqueue, queues_slots, succ_state);
				}
			   		
			}

		     }
		}
	    t = 1 - t;             
            

            __syncthreads();
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            

	    __syncthreads();     
     } while (!empty);
	for(i=0;i<49100;i++)
	{
	table[i] = s_table[i];
	}
}











//-----
//  With rep table, AB mix, gen32
//-----
__global__ void stateSpaceExplorationAB32BitEven (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a, unsigned long long int mix_b)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
        unsigned long long int       mix_state=0;
	unsigned long long int unmix_a=mix_a;
	unsigned long long int unmix_b=mix_b;
	unsigned long long int randND=mix_a;
	unsigned long long int selection=0;
	int sel=0;
	int nd = 0;
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
	int rep_value=0;
        //int unmixed=0;
	__shared__ uint8_t s_table[49100];
	uint8_t visitedState= 0;//not sure what all 0s is in char

        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];

	//BUG SOURCE
	//for(i=0;i<49100;i++)
	//{
	//s_table[i] = table[i];
	//}

	for (i = threadIdx.x; i < 49100; i += blockDim.x) 
    		s_table[i] = 0;
	__syncthreads();

 
     
	do 
        {
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
	
			for(iproc=0; iproc<numproc; iproc+=2)
			{
				nd=randND%4;
				succ_state=execBitGen32(iproc, nd, state);//calls the device function that returns a proper successor
				
				visitedState=0;//in char
				selection=0;//same

				mix_state=succ_state;
				mix_a=unmix_a;
				mix_b=unmix_b;
				selection=(mix(mix_a, mix_b, mix_state));
				hashed_value=(selection/8)%table_size;
				sel=selection%8;//to select bit
				randND=mix_a;

				//pull copy
				visitedState=s_table[hashed_value];  //this is what was hashed at this location (or 0 if nothing was here)
				
				s_table[hashed_value]|=(1<<sel);  //put our current value in the table (bitwise set)

				if((visitedState&(1<<sel))==0)//if this bit was not previously visited
				{
					//put generated successor in random "mailbox" for next round
					rep_value=hashed_value+(50000*sel);
					rep_table[rep_value]=succ_state;  //FOR REPORTING ONLY!  DISABLE ON REAL RUNS
					w_new = curand(&localState) % N;
					lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;
					insert_state_in_queue (lwqueue, queues_slots, succ_state);
				}
			   		
			}

		     }
		}
	    t = 1 - t;             
            

            __syncthreads();
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            

	    __syncthreads();     
     } while (!empty);
	for(i=0;i<49100;i++)
	{
	table[i] = s_table[i];
	}
}










//-----
//  With rep table, AB mix, gen32, all ND random order (inner loop)
//-----
__global__ void stateSpaceExplorationAB32BitNDInR (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a, unsigned long long int mix_b)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
        unsigned long long int       mix_state=0;
	unsigned long long int unmix_a=mix_a;
	unsigned long long int unmix_b=mix_b;
	unsigned long long int randND=mix_a;
	unsigned long long int selection=0;
	int sel=0;
	int nd = 0;
	int nd_sel=0;
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
	int rep_value=0;
        //int unmixed=0;
	__shared__ uint8_t s_table[49100];
	uint8_t visitedState= 0;//not sure what all 0s is in char

        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];

	//BUG SOURCE
	//for(i=0;i<49100;i++)
	//{
	//s_table[i] = table[i];
	//}

        nd_sel=randND%20;//setting random ND order

	for (i = threadIdx.x; i < 49100; i += blockDim.x) 
    		s_table[i] = 0;
	__syncthreads();

 
     
	do 
        {
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
			for(iproc=0; iproc<numproc; iproc++){
			for(nd=0; nd<4; nd++)
			{
				//nd=randND%4;
				succ_state=execBitGen32(iproc, consOrder[(nd_sel*4)+nd], state);//calls the device function that returns a proper successor
				
				visitedState=0;//in char
				selection=0;//same

				mix_state=succ_state;
				mix_a=unmix_a;
				mix_b=unmix_b;
				selection=(mix(mix_a, mix_b, mix_state));
				hashed_value=(selection/8)%table_size;
				sel=selection%8;//to select bit
				//randND=mix_a;

				//pull copy
				visitedState=s_table[hashed_value];  //this is what was hashed at this location (or 0 if nothing was here)
				
				s_table[hashed_value]|=(1<<sel);  //put our current value in the table (bitwise set)

				if((visitedState&(1<<sel))==0)//if this bit was not previously visited
				{
					//put generated successor in random "mailbox" for next round
					rep_value=hashed_value+(50000*sel);
					rep_table[rep_value]=succ_state;  //FOR REPORTING ONLY!  DISABLE ON REAL RUNS
					w_new = curand(&localState) % N;
					lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;
					if((lwqueue->stop) < 2*NxNxMAX_QUEUE_NUM_SLOTS)
					{
						insert_state_in_queue (lwqueue, queues_slots, succ_state);
					}
				}
			   		
			}}

		     }
		}
	    t = 1 - t;             
            

            __syncthreads();
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            

	    __syncthreads();     
     } while (!empty);
	for(i=0;i<49100;i++)
	{
	table[i] = s_table[i];
	}
}


















//-----
//  With rep table, AB mix, gen32, all nd random order (outer loop)
//-----
__global__ void stateSpaceExplorationAB32BitNDOutR (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a, unsigned long long int mix_b)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
        unsigned long long int       mix_state=0;
	unsigned long long int unmix_a=mix_a;
	unsigned long long int unmix_b=mix_b;
	unsigned long long int randND=mix_a;
	unsigned long long int selection=0;
	int sel=0;
	int nd = 0;
	int nd_sel=0;//random ND ordering
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
	int rep_value=0;
        //int unmixed=0;
	__shared__ uint8_t s_table[49100];
	uint8_t visitedState= 0;//not sure what all 0s is in char

        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];

	//BUG SOURCE
	//for(i=0;i<49100;i++)
	//{
	//s_table[i] = table[i];
	//}

        nd_sel=randND%20;//setting random ND order

	for (i = threadIdx.x; i < 49100; i += blockDim.x) 
    		s_table[i] = 0;
	__syncthreads();

 
     
	do 
        {
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
			for(nd=0;nd<4;nd++){
			for(iproc=0; iproc<numproc; iproc++)
			{
				//nd=randND%4;
				succ_state=execBitGen32(iproc, consOrder[(nd_sel*4)+nd], state);//calls the device function that returns a proper successor
				
				visitedState=0;//in char
				selection=0;//same

				mix_state=succ_state;
				mix_a=unmix_a;
				mix_b=unmix_b;
				selection=(mix(mix_a, mix_b, mix_state));
				hashed_value=(selection/8)%table_size;
				sel=selection%8;//to select bit
				//randND=mix_a;

				//pull copy
				visitedState=s_table[hashed_value];  //this is what was hashed at this location (or 0 if nothing was here)
				
				s_table[hashed_value]|=(1<<sel);  //put our current value in the table (bitwise set)

				if((visitedState&(1<<sel))==0)//if this bit was not previously visited
				{
					//put generated successor in random "mailbox" for next round
					rep_value=hashed_value+(50000*sel);
					rep_table[rep_value]=succ_state;  //FOR REPORTING ONLY!  DISABLE ON REAL RUNS
					w_new = curand(&localState) % N;
					lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;
					insert_state_in_queue (lwqueue, queues_slots, succ_state);
				}
			   		
			}}

		     }
		}
	    t = 1 - t;             
            

            __syncthreads();
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            

	    __syncthreads();     
     } while (!empty);
	for(i=0;i<49100;i++)
	{
	table[i] = s_table[i];
	}
}
















//-----
//  With rep table, AB mix, gen32, all ND standard order (inner loop)
//-----
__global__ void stateSpaceExplorationAB32BitNDIn (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a, unsigned long long int mix_b)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
        unsigned long long int       mix_state=0;
	unsigned long long int unmix_a=mix_a;
	unsigned long long int unmix_b=mix_b;
	//unsigned long long int randND=mix_a;
	unsigned long long int selection=0;
	int sel=0;
	int nd = 0;
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
	int rep_value=0;
        //int unmixed=0;
	__shared__ uint8_t s_table[49100];
	uint8_t visitedState= 0;//not sure what all 0s is in char

        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];

	//BUG SOURCE
	//for(i=0;i<49100;i++)
	//{
	//s_table[i] = table[i];
	//}

	for (i = threadIdx.x; i < 49100; i += blockDim.x) 
    		s_table[i] = 0;
	__syncthreads();

 
     
	do 
        {
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
			for(iproc=0; iproc<numproc; iproc++){
			for(nd=0; nd<4; nd++)
			{
				//nd=randND%4;
				succ_state=execBitGen32(iproc, nd, state);//calls the device function that returns a proper successor
				
				visitedState=0;//in char
				selection=0;//same

				mix_state=succ_state;
				mix_a=unmix_a;
				mix_b=unmix_b;
				selection=(mix(mix_a, mix_b, mix_state));
				hashed_value=(selection/8)%table_size;
				sel=selection%8;//to select bit
				//randND=mix_a;

				//pull copy
				visitedState=s_table[hashed_value];  //this is what was hashed at this location (or 0 if nothing was here)
				
				s_table[hashed_value]|=(1<<sel);  //put our current value in the table (bitwise set)

				if((visitedState&(1<<sel))==0)//if this bit was not previously visited
				{
					//put generated successor in random "mailbox" for next round
					rep_value=hashed_value+(50000*sel);
					rep_table[rep_value]=succ_state;  //FOR REPORTING ONLY!  DISABLE ON REAL RUNS
					w_new = curand(&localState) % N;
					lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;
					insert_state_in_queue (lwqueue, queues_slots, succ_state);
				}
			   		
			}}

		     }
		}
	    t = 1 - t;             
            

            __syncthreads();
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            

	    __syncthreads();     
     } while (!empty);
	for(i=0;i<49100;i++)
	{
	table[i] = s_table[i];
	}
}


















//-----
//  With rep table, AB mix, gen32, all nd standard order (outer loop)
//-----
__global__ void stateSpaceExplorationAB32BitNDOut (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a, unsigned long long int mix_b)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
        unsigned long long int       mix_state=0;
	unsigned long long int unmix_a=mix_a;
	unsigned long long int unmix_b=mix_b;
	//unsigned long long int randND=mix_a;
	unsigned long long int selection=0;
	int sel=0;
	int nd = 0;
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
	int rep_value=0;
        //int unmixed=0;
	__shared__ uint8_t s_table[49100];
	uint8_t visitedState= 0;//not sure what all 0s is in char

        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];

	//BUG SOURCE
	//for(i=0;i<49100;i++)
	//{
	//s_table[i] = table[i];
	//}

	for (i = threadIdx.x; i < 49100; i += blockDim.x) 
    		s_table[i] = 0;
	__syncthreads();

 
     
	do 
        {
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
			for(nd=0;nd<4;nd++){
			for(iproc=0; iproc<numproc; iproc++)
			{
				//nd=randND%4;
				succ_state=execBitGen32(iproc, nd, state);//calls the device function that returns a proper successor
				
				visitedState=0;//in char
				selection=0;//same

				mix_state=succ_state;
				mix_a=unmix_a;
				mix_b=unmix_b;
				selection=(mix(mix_a, mix_b, mix_state));
				hashed_value=(selection/8)%table_size;
				sel=selection%8;//to select bit
				//randND=mix_a;

				//pull copy
				visitedState=s_table[hashed_value];  //this is what was hashed at this location (or 0 if nothing was here)
				
				s_table[hashed_value]|=(1<<sel);  //put our current value in the table (bitwise set)

				if((visitedState&(1<<sel))==0)//if this bit was not previously visited
				{
					//put generated successor in random "mailbox" for next round
					rep_value=hashed_value+(50000*sel);
					rep_table[rep_value]=succ_state;  //FOR REPORTING ONLY!  DISABLE ON REAL RUNS
					w_new = curand(&localState) % N;
					lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;
					insert_state_in_queue (lwqueue, queues_slots, succ_state);
				}
			   		
			}}

		     }
		}
	    t = 1 - t;             
            

            __syncthreads();
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            

	    __syncthreads();     
     } while (!empty);
	for(i=0;i<49100;i++)
	{
	table[i] = s_table[i];
	}
}





//-----
//  With rep table, AB mix, gen32
//-----
__global__ void stateSpaceExplorationAB32Bit (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a, unsigned long long int mix_b)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
        unsigned long long int       mix_state=0;
	unsigned long long int unmix_a=mix_a;
	unsigned long long int unmix_b=mix_b;
	unsigned long long int randND=mix_a;
	unsigned long long int selection=0;
	int sel=0;
	int nd = 0;
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
	int rep_value=0;
        //int unmixed=0;
	__shared__ uint8_t s_table[49100];
	uint8_t visitedState= 0;//not sure what all 0s is in char

        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];

	//BUG SOURCE
	//for(i=0;i<49100;i++)
	//{
	//s_table[i] = table[i];
	//}

	for (i = threadIdx.x; i < 49100; i += blockDim.x) 
    		s_table[i] = 0;
	__syncthreads();

 
     
	do 
        {
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
	
			for(iproc=0; iproc<numproc; iproc++)
			{
				nd=randND%4;
				succ_state=execBitGen32(iproc, nd, state);//calls the device function that returns a proper successor
				
				visitedState=0;//in char
				selection=0;//same

				mix_state=succ_state;
				mix_a=unmix_a;
				mix_b=unmix_b;
				selection=(mix(mix_a, mix_b, mix_state));
				hashed_value=(selection/8)%table_size;
				sel=selection%8;//to select bit
				randND=mix_a;

				//pull copy
				visitedState=s_table[hashed_value];  //this is what was hashed at this location (or 0 if nothing was here)
				
				s_table[hashed_value]|=(1<<sel);  //put our current value in the table (bitwise set)

				if((visitedState&(1<<sel))==0)//if this bit was not previously visited
				{
					//put generated successor in random "mailbox" for next round
					rep_value=hashed_value+(50000*sel);
					rep_table[rep_value]=succ_state;  //FOR REPORTING ONLY!  DISABLE ON REAL RUNS
					w_new = curand(&localState) % N;
					lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;
					if((lwqueue->stop) < 2*NxNxMAX_QUEUE_NUM_SLOTS)
					{
					    insert_state_in_queue (lwqueue, queues_slots, succ_state);
					}
				}
			   		
			}

		     }
		}
	    t = 1 - t;             
            

            __syncthreads();
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            

	    __syncthreads();     
     } while (!empty);
	for(i=0;i<49100;i++)
	{
	table[i] = s_table[i];
	}
}
















//-----
//  With rep table, AB mix, gen32
//-----
__global__ void stateSpaceExplorationAB32S (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				bool * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a, unsigned long long int mix_b)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
        unsigned long long int       mix_state=0;
	unsigned long long int unmix_a=mix_a;
	unsigned long long int unmix_b=mix_b;
	int nd = 0;
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
        //int unmixed=0;
	__shared__ bool s_table[49100];
	bool visitedState= false;

        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];
	for(i=0;i<49100;i++)
	{
	s_table[i] = table[i];
	}

 
     
	do 
        {
		for (nd=0; nd<4; nd++){
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
	
			for(iproc=0; iproc<numproc; iproc++)
			{
				
				succ_state=execBitGen32(iproc, nd, state);//calls the device function that returns a proper successor
				
				visitedState=false;
				//unmixed=hash_function(function_set, 0, succ_state, table_size);
				mix_state=succ_state;
				mix_a=unmix_a;
				mix_b=unmix_b;
				hashed_value=((mix(mix_a, mix_b, mix_state))%table_size);

				//visitedState=add(hashed_value, table);
				visitedState=s_table[hashed_value];  //this is what was hashed at this location (or 0 if nothing was here)
				s_table[hashed_value]=1;  //put our current value in the table

				if(!visitedState)
				{
					//put generated successor in random "mailbox" for next round
					rep_table[hashed_value]=succ_state;  //FOR REPORTING ONLY!  DISABLE ON REAL RUNS
					w_new = curand(&localState) % N;
					lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;
					insert_state_in_queue (lwqueue, queues_slots, succ_state);
				}
			   		
			}

		     }
		}
	    t = 1 - t;             
            

            __syncthreads();
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            }

	    __syncthreads();     
     } while (!empty);
	for(i=0;i<49100;i++)
	{
	table[i] = s_table[i];
	}
}




//-----
//  With rep table, AB mix, gen32
//-----
__global__ void stateSpaceExplorationAB32 (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				bool * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a, unsigned long long int mix_b)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
        unsigned long long int       mix_state=0;
	unsigned long long int unmix_a=mix_a;
	unsigned long long int unmix_b=mix_b;
	unsigned long long int randND=mix_a;
	int nd = 0;
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
        //int unmixed=0;
	__shared__ bool s_table[49100];
	bool visitedState= false;

        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];
	for(i=0;i<49100;i++)
	{
	s_table[i] = table[i];
	}

 
     
	do 
        {
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
	
			for(iproc=0; iproc<numproc; iproc++)
			{
				nd=randND%4;
				succ_state=execBitGen32(iproc, nd, state);//calls the device function that returns a proper successor
				
				visitedState=false;
				//unmixed=hash_function(function_set, 0, succ_state, table_size);
				mix_state=succ_state;
				mix_a=unmix_a;
				mix_b=unmix_b;
				hashed_value=((mix(mix_a, mix_b, mix_state))%table_size);
				randND=mix_a;
				//visitedState=add(hashed_value, table);
				visitedState=s_table[hashed_value];  //this is what was hashed at this location (or 0 if nothing was here)
				s_table[hashed_value]=1;  //put our current value in the table

				if(!visitedState)
				{
					//put generated successor in random "mailbox" for next round
					rep_table[hashed_value]=succ_state;  //FOR REPORTING ONLY!  DISABLE ON REAL RUNS
					w_new = curand(&localState) % N;
					lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;
					insert_state_in_queue (lwqueue, queues_slots, succ_state);
				}
			   		
			}

		     }
		}
	    t = 1 - t;             
            

            __syncthreads();
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            

	    __syncthreads();     
     } while (!empty);
	for(i=0;i<49100;i++)
	{
	table[i] = s_table[i];
	}
}









//-----
//  With rep table
//-----
__global__ void stateSpaceExplorationAB (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				bool * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a, unsigned long long int mix_b)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
        unsigned long long int       mix_state=0;
	unsigned long long int unmix_a=mix_a;
	unsigned long long int unmix_b=mix_b;
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
        //int unmixed=0;
	__shared__ bool s_table[49100];
	bool visitedState= false;

        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];
	for(i=0;i<49100;i++)
	{
	s_table[i] = table[i];
	}

 
     
	do 
        {
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
	
			for(iproc=0; iproc<numproc; iproc++)
			{
				
				succ_state=execBit(iproc, numproc, state);//calls the device function that returns a proper successor
				
				visitedState=false;
				//unmixed=hash_function(function_set, 0, succ_state, table_size);
				mix_state=succ_state;
				mix_a=unmix_a;
				mix_b=unmix_b;
				hashed_value=((mix(mix_a, mix_b, mix_state))%table_size);
				//visitedState=add(hashed_value, table);
				visitedState=s_table[hashed_value];  //this is what was hashed at this location (or 0 if nothing was here)
				s_table[hashed_value]=1;  //put our current value in the table

				if(!visitedState)
				{
					//put generated successor in random "mailbox" for next round
					rep_table[hashed_value]=succ_state;  //FOR REPORTING ONLY!  DISABLE ON REAL RUNS
					w_new = curand(&localState) % N;
					lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;
					insert_state_in_queue (lwqueue, queues_slots, succ_state);
				}
			   		
			}

		     }
		}
	    t = 1 - t;             
            

            __syncthreads();
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            

	    __syncthreads();     
     } while (!empty);
	for(i=0;i<49100;i++)
	{
	table[i] = s_table[i];
	}
}





//-----
//  With rep table
//-----
__global__ void stateSpaceExploration (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				bool * table,
				int table_size, unsigned long long int * rep_table)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
	__shared__ bool s_table[49100];
	bool visitedState= false;

        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];
	for(i=0;i<49100;i++)
	{
	s_table[i] = table[i];
	}

 
     
	do 
        {
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
	
			for(iproc=0; iproc<numproc; iproc++)
			{
				
				succ_state=execBit(iproc, numproc, state);//calls the device function that returns a proper successor
				
				visitedState=false;
				hashed_value=hash_function(function_set, 0, succ_state, table_size);
				//visitedState=add(hashed_value, table);
				visitedState=s_table[hashed_value];  //this is what was hashed at this location (or 0 if nothing was here)
				s_table[hashed_value]=1;  //put our current value in the table

				if(!visitedState)
				{
					//put generated successor in random "mailbox" for next round
					rep_table[hashed_value]=succ_state;  //FOR REPORTING ONLY!  DISABLE ON REAL RUNS
					w_new = curand(&localState) % N;
					lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;
					insert_state_in_queue (lwqueue, queues_slots, succ_state);
				}
			   		
			}

		     }
		}
	    t = 1 - t;             
            

            __syncthreads();
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            

	    __syncthreads();     
     } while (!empty);
	for(i=0;i<49100;i++)
	{
	table[i] = s_table[i];
	}
}


//-----
// Normal
//-----

/*__global__ void stateSpaceExploration (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				bool * table,
				int table_size)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
	__shared__ bool s_table[49100];
	bool visitedState= false;

        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];
	for(i=0;i<49100;i++)
	{
	s_table[i] = table[i];
	}

 
     
	do 
        {
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
	
			for(iproc=0; iproc<numproc; iproc++)
			{
				
				succ_state=execBit(iproc, numproc, state);//calls the device function that returns a proper successor
				
				visitedState=false;
				hashed_value=hash_function(function_set, 0, succ_state, table_size);
				//visitedState=add(hashed_value, table);
				visitedState=s_table[hashed_value];  //this is what was hashed at this location (or 0 if nothing was here)
				s_table[hashed_value]=1;  //put our current value in the table

				if(!visitedState)
				{
					//put generated successor in random "mailbox" for next round
					w_new = curand(&localState) % N;
					lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;
					insert_state_in_queue (lwqueue, queues_slots, succ_state);
				}
			   		
			}

		     }
		}
	    t = 1 - t;             
            

            __syncthreads();
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            

	    __syncthreads();     
     } while (!empty);
	for(i=0;i<49100;i++)
	{
	table[i] = s_table[i];
	}
}*/






//-----
//  With non-determinism
//-----


/*
__global__ void stateSpaceExplorationND (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states, 
				int numstates,
                        curandState * cstate,
				unsigned long long int * table,
				unsigned long long int * stashtable,
				int table_size,
				volatile int * gpu_ArrayIn, 
                                volatile int * gpu_ArrayOut)
{

        Queue   * lqueue;
        Queue   * lwqueue;
        int       w,q,w_new,j,z,i;
        unsigned long long int       state;
	unsigned long long int       succ_state;
	unsigned long long int       evic_state=1;//must start at 1 because 0 stops the while loop
        //unsigned long long int       devic_state=1;//used as part of an unnecessary check
	int     t = 0;
	int	iproc;
	int	numproc=numstates;//right now is number of philosophers, but could be any process	
	int hashed_value=0;
	int chain=0;
	bool visitedState= false;
        int nd=0;


        w     = blockIdx.x * blockDim.x + threadIdx.x; // Search(w: 1..N)     Workers
        curandState localState = cstate[w];

 
     
	do 
        {
		for (q=0; q < N; q++)
		{
		     lqueue       = cores_queues + t*NxN + w*N + q;
		     while ((state=extract_state_from_queue(lqueue, queues_slots)) != 18446744073709551615)//because unsigned
		     {
			for (nd=0; nd<2; nd++)
			{
	
			for(iproc=0; iproc<numproc; iproc++)
			{
				
				succ_state=execBitLamport3((long long int)iproc, (long long int)numproc, state);//calls the device function that returns a proper successor
	//if(function_set->constants[NUMHASHFUNC].x==0){succ_state=11111;}			
				i=0;
				visitedState=false;
				visitedState=checkStash(succ_state, function_set->constants[NUMHASHFUNC], stashtable, table_size);
				while(i<NUMHASHFUNC)  //check if we have seen this state before
			    	{
					if(visitedState){break;}//we have, stop
					visitedState=checkState(i, succ_state, function_set, table, table_size);
					if(visitedState){break;}//we have, stop
					i++;
			    	}
			    	if (i==NUMHASHFUNC)//inelegant way to suggest we only do this if we have not already visited this state
			    	{
					hashed_value=hash_function(function_set, 0, succ_state, table_size);//put it in its first feasible slot

					//property check goes here, breaking if violation is found
					evic_state=add(hashed_value, succ_state, table);  //add global state vector to hash table, return evicted value (possibly 0)
					//evic_state=(evic_state == succ_state ? 0 : evic_state);//a nice check, but doesn't do anything useful in practice

					while(evic_state!=0 && chain<MAXITERATIONS)
					{
				  		j=0;//yes, I have co-opted j from the sync, but it should not interfere with that
				  		j=findState(hashed_value, evic_state, function_set, table, table_size);//this is the same hashed_value out of the last add() call
				  		hashed_value=hash_function(function_set, (j+1%NUMHASHFUNC), evic_state, table_size);//this is the incremented hashed_value
				  		evic_state = add(hashed_value,evic_state, table);
				  		chain++;
					}
					if(chain>=MAXITERATIONS)//Ideally this wouldn't be an if, but I digress
					{
				  		hashed_value=stash_hash_function(function_set->constants[NUMHASHFUNC], evic_state, table_size);//need to call stash function
				  		evic_state = stashadd(hashed_value, evic_state, stashtable);
						//evic_state = (devic_state == evic_state ? 0 : devic_state); //a nice check, but doesn't do anything useful in practice
					}
					chain=0;
					if(evic_state!=0){}//if this happens, we need to tune our hash functions

					//put generated successor in random "mailbox" for next round
					w_new = curand(&localState) % N;
					lwqueue = cores_queues + (1-t)*NxN + w_new*N + w;
					insert_state_in_queue (lwqueue, queues_slots, succ_state);


			   	}	
			}
			}
		     }
		}
	    t = 1 - t;             
            
            gpu_lock_free_sync(10, gpu_ArrayIn, gpu_ArrayOut);
            
            if (w == 0)
            {
                empty = 1;
               
                for (z=0; z < N; z++)
                {
                    for (j=0; j < N; j++)
                    {
                        lwqueue       = cores_queues + t*NxN + z*N + j;
                        if (lwqueue->start != lwqueue->stop)
                        {
                             empty = 0;
                             break;
                        } 
                    }
                    
                }
            }
            
            gpu_lock_free_sync(20, gpu_ArrayIn, gpu_ArrayOut);      
     } while (!empty);
}*/


//GPU lock-free synchronization function
__device__ void gpu_lock_free_sync(int goalVal, volatile int * Arrayin, volatile int * Arrayout)
{

    // thread ID in a block
    int tid_in_blk = threadIdx.x;
    
    int nBlockNum = gridDim.x;
    
    int bid = blockIdx.x;

    // only thread 0 is used for synchronization
   if (tid_in_blk == 0) {
       Arrayin[bid] = goalVal;
   }

   if (bid == 1) {
      if (tid_in_blk < nBlockNum) 
      {
       
          while (Arrayin[tid_in_blk] != goalVal)
          {
            //Do nothing here
          }
      }

      __syncthreads();

      if (tid_in_blk < nBlockNum) 
      {
         Arrayout[tid_in_blk] = goalVal;
      }
   }

   if (tid_in_blk == 0) 
   {
       while (Arrayout[bid] != goalVal) 
       {
          //Do nothing here
       }
   }
   __syncthreads();
 }


//-----
// Evaluation of execution (with state vector)
//
// Updates values if executable; returns same number else
//
// This looks at the CFG we made in main to determine a XOR map that would change the philosopher's
// local state and pertinent sticks if such a move is possible.  Other philosophers and sticks are
// unaffected by this XORing.  If no change is possible, then the original state vector is returned
// as the successor.
//
//
// given the current process, total number of processes, the current global state vector, and the control flow graph
// return the successor's global state vector
//-----
__device__ unsigned long long int execBit(int iproc, int numproc, unsigned long long int stVector)
{
	int CFGindex=0;
	int temp=0;
	int shorthand=(iproc+1)%numproc;
	unsigned long long int togglemask=0;
	
	CFGindex+= ( (unsigned long long int)((stVector) & (1<<(iproc*2))) >> (iproc*2));
	CFGindex+= ( (unsigned long long int)((stVector) & (1<<(iproc*2+1))) >> (iproc*2));
	CFGindex+= ( (unsigned long long int)((stVector) & ((unsigned long long int)1<<(iproc+(numproc*2)))) >> (iproc+(numproc*2)-2));
	CFGindex+= ( (unsigned long long int)((stVector) & ((unsigned long long int)1<<((numproc*2)+shorthand))) >> ((numproc*2)+shorthand)-3);
	
	temp= consCFG[CFGindex];

	togglemask+= ( (unsigned long long int)((temp) & (1<<(0))) << (iproc*2));
	togglemask+= ( (unsigned long long int)((temp) & (1<<(1))) << (iproc*2));	
	togglemask+= ( (unsigned long long int)((temp) & (1<<(2))) << (iproc+(numproc*2))-2);
	togglemask+= ( (unsigned long long int)((temp) & (1<<(3))) << ((numproc*2)+shorthand)-3);
	
	stVector ^= togglemask;

	return (stVector);
}


//-----
// Evaluation of execution (with state vector) for Leader Election
//
// Updates values if follower; returns same number else
//
//
//
//
//-----
/*__device__ unsigned long long int execBit(int iproc, int numproc, unsigned long long int stVector)
{
	
        int                    shorthand  = (iproc+1) % numproc;
        unsigned long long int bit_mask   = 255;
        int                    ishift     = (iproc     * 8);
        int                    nshift     = (shorthand * 8);
        unsigned long long int ival       = (stVector & (bit_mask << ishift)) >> ishift;
        unsigned long long int nval       = (stVector & (bit_mask << nshift)) >> nshift;
        int CFGindex                      = (ival>=nval);
        //stVector                          = (stVector & (~(bit_mask << ishift))) + (CFGindex * (ival << ishift)) + ((1-CFGindex) * (nval << ishift));
	return (stVector + ((1-CFGindex) * ((nval-ival) << ishift)));
}*/


__device__ unsigned long long int execBitGen32 (long long int pid, int nd, unsigned long long int stVector)
{
	
	/*
	 int val;

	 inline check(){
		if
		:: (val == -1196372740)
		|| (val == -222966779)
		...
		  -> c_code { printf("assertion violated %d\n", now.val);}
		:: else //no match
		fi
	 }
	 active[8] proctype word()
	 { //_pid = 0..7 -- each proc owns 4 bits
	 end:   do
		:: d_step { val=val|1<<((4*_pid)+0); check}
		:: d_step { val=val|1<<((4*_pid)+1); check}
		:: d_step { val=val|1<<((4*_pid)+2); check}
		:: d_step { val=val|1<<((4*_pid)+3); check}
		od
	 }
	 */
    
    int mask = 0;
    long long int MAXINT = 4294967295;
    long long int barrier = 0;
    int temp_vector=((int)stVector);
    barrier = MAXINT - stVector;
    mask += (1<<((4*pid)+nd));


    //stVector += (barrier>=mask) * mask;
    stVector = stVector|mask;//more correct
    //temp_vector = temp_vector|mask;
    //stVector=temp_vector%2147483647;
    stVector=stVector%MAXINT;

    return stVector;
}


__device__ unsigned long long int execBitPeterson3 (long long int pid, long long int nd, long long int numproc, unsigned long long int stVector)
{
	
	/*
	 byte pos[3];
	 byte step[3];
	 
	 
	 active proctype P_0() {
	 byte j=0;
	 byte k=0;
	 
	 NCS: if
	 :: j = 1; goto wait;
	 
	 fi;
	 CS: if
	 :: pos[pid] = 0; goto NCS;
	 
	 fi;
	 wait: if
	 ::  d_step {j<numproc;pos[pid] = j;}  goto q2;
	 
	 :: j==numproc; goto CS;
	 
	 fi;
	 q2: if
	 ::  d_step {step[j-1] = 0;k = 0;}  goto q3;
	 
	 fi;
	 q3: if
	 ::  d_step {k<numproc && (k==pid || pos[k]<j);k = k+1;}  goto q3;
	 
	 ::  d_step {pos[j-1]!=pid || k==numproc;j = j+1;}  goto wait; 
	 
	 fi;
	 }
	 */
    
    
	
    long long int  bit_shift      = pid * PET_STATE_BITSHIFT;
    
    
    //Ezio: I extract the program counter and other important variables
    
    long long int               pc   = (stVector & (PET_PC_MASK   << bit_shift)) >>  bit_shift;
    long long int               pos  = (stVector & (PET_POS_MASK  << bit_shift)) >>  (bit_shift + PET_OFFSET_POS);
    long long int               step = (stVector & (PET_STEP_MASK << bit_shift)) >>  (bit_shift + PET_OFFSET_STEP);
    long long int                  j = (stVector & (PET_J_MASK    << bit_shift)) >>  (bit_shift + PET_OFFSET_J);
    long long int                  k = (stVector & (PET_K_MASK    << bit_shift)) >>  (bit_shift + PET_OFFSET_K);
    
    long long int          j_m_shift   = (j-1) * PET_STATE_BITSHIFT;
	long long int            k_shift   = k     * PET_STATE_BITSHIFT;
    
    long long int            step_j_m = (stVector & (PET_STEP_MASK  << j_m_shift)) >>  (j_m_shift + PET_OFFSET_STEP);
	
	long long int             pos_j_m = (stVector & (PET_POS_MASK  << j_m_shift)) >>  (j_m_shift + PET_OFFSET_POS);
	
    long long int               pos_k = (stVector & (PET_POS_MASK   << k_shift))   >>  (k_shift   + PET_OFFSET_POS);
    
    
    long long int pc_inc             = (1L << bit_shift);   
	long long int  k_inc             = (1L << (bit_shift + PET_OFFSET_K)); 
	long long int  j_inc             = (1L << (bit_shift + PET_OFFSET_J)); 
	
    //NCS: if
    //:: j = 1; goto wait;
    
    //fi;
	
	//printf("j=%lld\n", j);
	//printf("j_inc=%lld\n", j_inc);
    stVector += PET_NCS * (pc_inc + pc_inc + ((1L-j) << (bit_shift + PET_OFFSET_J)) );
    
	
	
    //CS: if
    //:: pos[pid] = 0; goto NCS;
    
    //fi;
    stVector += PET_CS * (-pc_inc - (pos << (bit_shift + PET_OFFSET_POS)));
	
	
    //wait: if
    //::  d_step {j<numproc;pos[pid] = j;}  goto q2;
    
    //:: j==numproc; goto CS;
    
    //fi;
	//printf("j=%lld\n", j);
	//printf("numproc=%lld\n", numproc);
    stVector += PET_WAIT * ((j<numproc) * (pc_inc + ((j-pos) << (bit_shift + PET_OFFSET_POS))) - (j==numproc) * pc_inc);
	
	
    //q2: if
    //::  d_step {step[j-1] = pid;k = 0;}  goto q3;
    
    //fi;
    stVector += PET_Q2 * (pc_inc - (k << (bit_shift + PET_OFFSET_K)) - (step_j_m << (j_m_shift + PET_OFFSET_STEP) ) + (pid << (j_m_shift + PET_OFFSET_STEP) ));
    
    
	
	
	
    //q3: if
    //::  d_step {k<3 && (k==pid || pos[k]<j);k = k+1;}  goto q3;
    
    //::  d_step {pos[j-1]!=pid || k==numproc;j = j+1;}  goto wait;
    
    //fi;
	stVector += PET_Q3 * ( (1-nd) * (k<numproc) * (k==pid || pos_k<j) * k_inc + nd * (pos_j_m!=pid || k==numproc) * (j_inc - 2L * pc_inc)   );
	
	
    return stVector;
}




__device__ unsigned long long int execBitPeterson2 (long long int pid, long long int nd, long long int numproc, unsigned long long int stVector)
{
	
	/*
	 byte pos[3];
	 byte step[3];
	 
	 
	 active proctype P_0() {
	 byte j=0;
	 byte k=0;
	 
	 NCS: if
	 :: j = 1; goto wait;
	 
	 fi;
	 CS: if
	 :: pos[pid] = 0; goto NCS;
	 
	 fi;
	 wait: if
	 ::  d_step {j<numproc;pos[pid] = j;}  goto q2;
	 
	 :: j==numproc; goto CS;
	 
	 fi;
	 q2: if
	 ::  d_step {step[j-1] = 0;k = 0;}  goto q3;
	 
	 fi;
	 q3: if
	 ::  d_step {k<numproc && (k==pid || pos[k]<=j);k = k+1;}  goto q3;
	 
	 ::  d_step {step[j-1]!=pid || k==numproc;j = j+1;}  goto wait; 
	 
	 fi;
	 }
	 */
    
    
	
    long long int  bit_shift      = pid * PET_STATE_BITSHIFT;
    
    
    //Ezio: I extract the program counter and other important variables
    
    long long int               pc   = (stVector & (PET_PC_MASK   << bit_shift)) >>  bit_shift;
    long long int               pos  = (stVector & (PET_POS_MASK  << bit_shift)) >>  (bit_shift + PET_OFFSET_POS);
    long long int               step = (stVector & (PET_STEP_MASK << bit_shift)) >>  (bit_shift + PET_OFFSET_STEP);
    long long int                  j = (stVector & (PET_J_MASK    << bit_shift)) >>  (bit_shift + PET_OFFSET_J);
    long long int                  k = (stVector & (PET_K_MASK    << bit_shift)) >>  (bit_shift + PET_OFFSET_K);
    
    long long int          j_m_shift   = (j-1) * PET_STATE_BITSHIFT;
	long long int            k_shift   = k     * PET_STATE_BITSHIFT;
    
    long long int            step_j_m = (stVector & (PET_STEP_MASK  << j_m_shift)) >>  (j_m_shift + PET_OFFSET_STEP);
    long long int               pos_k = (stVector & (PET_POS_MASK   << k_shift))   >>  (k_shift   + PET_OFFSET_POS);
    
    
    long long int pc_inc             = (1L << bit_shift);   
	long long int  k_inc             = (1L << (bit_shift + PET_OFFSET_K)); 
	long long int  j_inc             = (1L << (bit_shift + PET_OFFSET_J)); 
	
	long long int cond1,cond2;
	
    //NCS: if
    //:: j = 1; goto wait;
    
    //fi;
	
	//printf("j=%lld\n", j);
	//printf("j_inc=%lld\n", j_inc);
    stVector += PET_NCS * (pc_inc + pc_inc + ((1L-j) << (bit_shift + PET_OFFSET_J)) );
    
	
	
    //CS: if
    //:: pos[pid] = 0; goto NCS;
    
    //fi;
    stVector += PET_CS * (-pc_inc - (pos << (bit_shift + PET_OFFSET_POS)));
	
	
    //wait: if
    //::  d_step {j<numproc;pos[pid] = j;}  goto q2;
    
    //:: j==numproc; goto CS;
    
	cond1 = (j<numproc);
	cond2 = (j==numproc);
	
    //fi;
	//printf("j=%lld\n", j);
	//printf("numproc=%lld\n", numproc);
    stVector += PET_WAIT * (cond1 * (pc_inc + ((j-pos) << (bit_shift + PET_OFFSET_POS))) - cond2 * pc_inc);
	
	

    //q2: if
    //::  d_step {step[j-1] = pid;k = 0;}  goto q3;
    
    //fi;
    stVector += PET_Q2 * (pc_inc - (k << (bit_shift + PET_OFFSET_K)) - (step_j_m << (j_m_shift + PET_OFFSET_STEP) ) + (pid << (j_m_shift + PET_OFFSET_STEP) ));
    
    
	
	cond1 = (long long int)(k<numproc) * (long long int)(k==pid || pos_k<=j);
	cond2 = (long long int)(step_j_m!=pid || k==numproc);
	
    //q3: if
    //::  d_step {k<3 && (k==pid || pos[k]<=j);k = k+1;}  goto q3;
    
    //::  d_step {step[j-1]!=pid || k==numproc;j = j+1;}  goto wait;
    
    //fi;
	stVector += PET_Q3 * ( (1-nd) * cond1 * k_inc + nd * cond2 * (j_inc - 2L * pc_inc)   );
	
	
    return stVector;
}

//Peterson1
__device__ unsigned long long int execBitPeterson1 (long long int pid, long long int nd, long long int numproc, unsigned long long int stVector)
{
	/*
	 byte pos[3];
	 byte step[3];
	 
	 
	 active proctype P_0() {
	 byte j=0;
	 byte k=0;
	 
	 NCS: if
	 :: j = 1; goto wait;
	 
	 fi;
	 CS: if
	 :: pos[pid] = 0; goto NCS;
	 
	 fi;
	 wait: if
	 ::  d_step {j<numproc;pos[pid] = j;}  goto q2;
	 
	 :: j==numproc; goto CS;
	 
	 fi;
	 q2: if
	 ::  d_step {step[j-1] = 0;k = 0;}  goto q3;
	 
	 fi;
	 q3: if
	 ::  d_step {k<numproc && (k==pid || pos[k]<j);k = k+1;}  goto q3;
	 
	 ::  d_step {step[j-1]!=pid || k==numproc;j = j+1;}  goto wait; 
	 
	 fi;
	 }
	 */
    
    

    long long int  bit_shift      = pid * PET_STATE_BITSHIFT;
    
    
    //Ezio: I extract the program counter and other important variables
    
    long long int               pc   = (stVector & (PET_PC_MASK   << bit_shift)) >>  bit_shift;
    long long int               pos  = (stVector & (PET_POS_MASK  << bit_shift)) >>  (bit_shift + PET_OFFSET_POS);
    long long int               step = (stVector & (PET_STEP_MASK << bit_shift)) >>  (bit_shift + PET_OFFSET_STEP);
    long long int                  j = (stVector & (PET_J_MASK    << bit_shift)) >>  (bit_shift + PET_OFFSET_J);
    long long int                  k = (stVector & (PET_K_MASK    << bit_shift)) >>  (bit_shift + PET_OFFSET_K);
    
    long long int          j_m_shift   = (j-1) * PET_STATE_BITSHIFT;
	long long int            k_shift   = k     * PET_STATE_BITSHIFT;
    
    long long int            step_j_m = (stVector & (PET_STEP_MASK  << j_m_shift)) >>  (j_m_shift + PET_OFFSET_STEP);
    long long int               pos_k = (stVector & (PET_POS_MASK   << k_shift))   >>  (k_shift   + PET_OFFSET_POS);
    
    
    long long int pc_inc             = (1L << bit_shift);   
	long long int  k_inc             = (1L << (bit_shift + PET_OFFSET_K)); 
	long long int  j_inc             = (1L << (bit_shift + PET_OFFSET_J)); 

    //NCS: if
    //:: j = 1; goto wait;
    
    //fi;
	
	//printf("j=%lld\n", j);
	//printf("j_inc=%lld\n", j_inc);
    stVector += PET_NCS * (pc_inc + pc_inc + ((1L-j) << (bit_shift + PET_OFFSET_J)) );
    


    //CS: if
    //:: pos[pid] = 0; goto NCS;
    
    //fi;
    stVector += PET_CS * (-pc_inc - (pos << (bit_shift + PET_OFFSET_POS)));


    //wait: if
    //::  d_step {j<numproc;pos[pid] = j;}  goto q2;
    
    //:: j==numproc; goto CS;
    
    //fi;
	//printf("j=%lld\n", j);
	//printf("numproc=%lld\n", numproc);
    stVector += PET_WAIT * ((j<numproc) * (pc_inc + ((j-pos) << (bit_shift + PET_OFFSET_POS))) - (j==numproc) * pc_inc);


    //q2: if
    //::  d_step {step[j-1] = pid;k = 0;}  goto q3;
    
    //fi;
    stVector += PET_Q2 * (pc_inc - (k << (bit_shift + PET_OFFSET_K)) - (step_j_m << (j_m_shift + PET_OFFSET_STEP) ) + (pid << (j_m_shift + PET_OFFSET_STEP) ));
    
    
	
	
	
    //q3: if
    //::  d_step {k<3 && (k==pid || pos[k]<j);k = k+1;}  goto q3;
    
    //::  d_step {step[j-1]!=pid || k==numproc;j = j+1;}  goto wait;
    
    //fi;
	
	stVector += PET_Q3 * ( (1-nd) * (k<numproc) * (k==pid || pos_k<j) * k_inc + nd * (step_j_m!=pid || k==numproc) * (j_inc - 2L * pc_inc)   );
	

    return stVector;
}
































//Anderson Even
__device__ unsigned long long int execBitAE (int iproc, int numproc, unsigned long long int stVector)
{
    int pid=iproc;
    /* FROM BEEM MODELS
     
                  byte my_place=0;
    
        000:      NCS: if
                  ::  d_step {my_place = next;next = next+1;}  goto p1;
    
                  fi;
     
        001:      p1: if
                  ::  d_step {my_place==N-1;next = next-N;}  goto p2;
    
                  ::  d_step {my_place!=N-1;my_place = my_place%N;}  goto p2;
    
                  fi;
     
        010:      p2: if
                  :: Slot[my_place]==1; goto p3;
    
                  fi;
     
        011:      p3: if
                  :: Slot[my_place] = 0; goto CS;
    
                  fi;

        100:      CS: if
                  :: Slot[(my_place+1)%N] = 1; goto NCS;
    
                  fi;
                  }
    */
    
    /*unsigned*/ long long int  bit_shift      = pid * AND_STATE_BITSHIFT;
   
    
    //Ezio: I extract the program counter and other important variables
    
    /*unsigned*/ long long int               pc = (stVector & (AND_PC_MASK       << bit_shift)) >>  bit_shift;
    /*unsigned*/ long long int         my_place = (stVector & (AND_MYPLACE_MASK  << bit_shift)) >>  (bit_shift + AND_OFFSET_MYPLACE);
    
    /*unsigned*/ long long int    myplace_shift = my_place * AND_STATE_BITSHIFT;
    
    /*unsigned*/ long long int    slot_my_place = (stVector & (AND_SLOT_MASK     << myplace_shift)) >>  (myplace_shift + AND_OFFSET_SLOT);
    /*unsigned*/ long long int             next =  stVector >> AND_NEXTSHIFT;
    
    
    
    
    /*unsigned*/ long long int pc_inc = (1L << bit_shift);
    
    stVector += NCS * (pc_inc - (stVector & (AND_MYPLACE_MASK << bit_shift)) + (next << (bit_shift + AND_OFFSET_MYPLACE)) + (1L << (AND_NEXTSHIFT)));
    // NCS: if
    // ::  d_step {my_place = next;next = next+1;}  goto p1;
    // fi
    
    
    
    stVector -= P1  * (-pc_inc + ((my_place==numproc-1) * (((unsigned long long int)numproc) << AND_NEXTSHIFT))  + (my_place!=numproc-1) * ((my_place-(my_place%numproc)) << (bit_shift + AND_OFFSET_MYPLACE)));
    // p1: if
    // ::  d_step {my_place==N-1;next = next-N;}  goto p2;
    // ::  d_step {my_place!=N-1;my_place = my_place%N;}  goto p2;
    // fi
    stVector += P2  * ((slot_my_place==1) * pc_inc);
    // p2: if
    // :: Slot[my_place]==1; goto p3;
    // fi
    stVector += P3  * (pc_inc - (stVector & (AND_SLOT_MASK << (my_place * AND_STATE_BITSHIFT))));
    // p3: if
    // :: Slot[my_place] = 0; goto CS;even Slot[(my_place+N-1)%N] = 0; goto CS;odd
    // fi
    
     stVector += CS  * (-4 * pc_inc - (stVector & (AND_SLOT_MASK << (((my_place+1)%numproc) * AND_STATE_BITSHIFT))) + (1L << (( ((my_place+1)%numproc) * AND_STATE_BITSHIFT) + AND_OFFSET_SLOT))); //- (stVector & (AND_SLOT_MASK << (((my_place+1)%numproc) * AND_STATE_BITSHIFT))) + (1L <<  (((my_place+1)%numproc) * AND_STATE_BITSHIFT) ));
    // CS: if
    // :: Slot[(my_place+1)%N] = 1; goto NCS;
    // fi
    
    //(stVector & (PET_F_MASK << bit_shift)) + (k << (bit_shift + PET_OFFSET_F))
    //stVector += (pc==0) * (pc_inc - (stVector & (PET_K_MASK << bit_shift)));
    
    
    return stVector;
}








//Anderson Odd
__device__ unsigned long long int execBitAO (int iproc, int numproc, unsigned long long int stVector)
{
    int pid=iproc;
    /* FROM BEEM MODELS
     
                  byte my_place=0;
    
        000:      NCS: if
                  ::  d_step {my_place = next;next = next+1;}  goto p1;
    
                  fi;
     
        001:      p1: if
                  ::  d_step {my_place==N-1;next = next-N;}  goto p2;
    
                  ::  d_step {my_place!=N-1;my_place = my_place%N;}  goto p2;
    
                  fi;
     
        010:      p2: if
                  :: Slot[my_place]==1; goto p3;
    
                  fi;
     
        011:      p3: if
                  :: Slot[my_place] = 0; goto CS;
    
                  fi;

        100:      CS: if
                  :: Slot[(my_place+1)%N] = 1; goto NCS;
    
                  fi;
                  }
    */
    
    /*unsigned*/ long long int  bit_shift      = pid * AND_STATE_BITSHIFT;
   
    
    //Ezio: I extract the program counter and other important variables
    
    /*unsigned*/ long long int               pc = (stVector & (AND_PC_MASK       << bit_shift)) >>  bit_shift;
    /*unsigned*/ long long int         my_place = (stVector & (AND_MYPLACE_MASK  << bit_shift)) >>  (bit_shift + AND_OFFSET_MYPLACE);
    
    /*unsigned*/ long long int    myplace_shift = my_place * AND_STATE_BITSHIFT;
    
    /*unsigned*/ long long int    slot_my_place = (stVector & (AND_SLOT_MASK     << myplace_shift)) >>  (myplace_shift + AND_OFFSET_SLOT);
    /*unsigned*/ long long int             next =  stVector >> AND_NEXTSHIFT;
    
    
    
    
    /*unsigned*/ long long int pc_inc = (1L << bit_shift);
    
    stVector += NCS * (pc_inc - (stVector & (AND_MYPLACE_MASK << bit_shift)) + (next << (bit_shift + AND_OFFSET_MYPLACE)) + (1L << (AND_NEXTSHIFT)));
    // NCS: if
    // ::  d_step {my_place = next;next = next+1;}  goto p1;
    // fi
    
    
    
    stVector -= P1  * (-pc_inc + ((my_place==numproc-1) * (((unsigned long long int)numproc) << AND_NEXTSHIFT))  + (my_place!=numproc-1) * ((my_place-(my_place%numproc)) << (bit_shift + AND_OFFSET_MYPLACE)));
    // p1: if
    // ::  d_step {my_place==N-1;next = next-N;}  goto p2;
    // ::  d_step {my_place!=N-1;my_place = my_place%N;}  goto p2;
    // fi
    stVector += P2  * ((slot_my_place==1) * pc_inc);
    // p2: if
    // :: Slot[my_place]==1; goto p3;
    // fi
    stVector += P3  * (pc_inc - (stVector & (AND_SLOT_MASK << (((my_place+numproc-1)%numproc) * AND_STATE_BITSHIFT))));
    // p3: if
    // :: Slot[my_place] = 0; goto CS;even Slot[(my_place+N-1)%N] = 0; goto CS;odd
    // fi
    
     stVector += CS  * (-4 * pc_inc - (stVector & (AND_SLOT_MASK << (((my_place+1)%numproc) * AND_STATE_BITSHIFT))) + (1L << (( ((my_place+1)%numproc) * AND_STATE_BITSHIFT) + AND_OFFSET_SLOT))); //- (stVector & (AND_SLOT_MASK << (((my_place+1)%numproc) * AND_STATE_BITSHIFT))) + (1L <<  (((my_place+1)%numproc) * AND_STATE_BITSHIFT) ));
    // CS: if
    // :: Slot[(my_place+1)%N] = 1; goto NCS;
    // fi
    
    //(stVector & (PET_F_MASK << bit_shift)) + (k << (bit_shift + PET_OFFSET_F))
    //stVector += (pc==0) * (pc_inc - (stVector & (PET_K_MASK << bit_shift)));
    
    
    return stVector;
}








__device__ unsigned long long int execBitLamport1 (long long int pid, long long int numproc, unsigned long long int stVector)
{

    /* byte b[3];
       byte x=255;
       byte y=255;


       active proctype P_0() {
           byte j=0;
    
       00: NCS: if
           :: b[pid] = 1; goto q1;
           fi;
     
       01: CS: if
           :: y = 255; goto e1;
    
           fi;

       02: q1: if
           :: x = pid; goto q2;
           fi;
        
       03: q2: if
           ::  d_step {y!=255;b[pid] = 0;}  goto q22;
    
           :: y==255; goto p;
    
           fi;
        04: q22: if
            :: y==255; goto NCS;
    
            fi;
            
        05: p: if
            :: y = pid; goto q3;
    
            fi;
            
        06: q3: if
            :: x==pid; goto CS;
    
            ::  d_step {x!=pid;b[pid] = 0;j = 0;}  goto q4;
    
            fi;
            
        07: q4: if
            ::  d_step {j<N && b[j]==0;j = j+1;}  goto q4;
    
            :: j==N; goto q5;
    
            fi;
            
        08: q5: if
            :: y==pid; goto CS;
    
            :: y==255; goto NCS;
    
            fi;
            
        09: e1: if
            :: b[pid] = 0; goto NCS;
    
            fi;
}*/

    long long int  bit_shift      = pid * LAMP_STATE_BITSHIFT;
    
    
    //Ezio: I extract the program counter and other important variables
    
    long long int               pc = (stVector & (LAMP_PC_MASK    << bit_shift)) >>  bit_shift;
    long long int                b = (stVector & (LAMP_SLOT_MASK  << bit_shift)) >>  (bit_shift + LAMP_OFFSET_SLOT);
    long long int                j = (stVector & (LAMP_J_MASK     << bit_shift)) >>  (bit_shift + LAMP_OFFSET_J);
    
    long long int          j_shift = j * LAMP_STATE_BITSHIFT;
    
    long long int              b_j = (stVector & (LAMP_SLOT_MASK  << j_shift)) >>  (j_shift + LAMP_OFFSET_SLOT);
    
    long long int             x    =  (stVector & LAMP_X_MASK) >> LAMP_X_SHIFT;
    long long int             y    =  (stVector & LAMP_Y_MASK) >> LAMP_Y_SHIFT;
    
    
    
    long long int pc_inc           = (1L << bit_shift);
	
	long long int cond1, cond2;

    // NCS: if
    //    :: b[pid] = 1; goto q1;
    //fi;
    stVector += LAMP_NCS * (2L *pc_inc +  ((1-b) << (bit_shift + LAMP_OFFSET_SLOT)));
    


    //CS: if
    //    :: y = 255; goto e1;
    //fi;
    stVector += LAMP_CS * (8L * pc_inc + ((LAMP_MAX_Y-y) << (LAMP_Y_SHIFT)));


    //q1: if
    //:: x = pid; goto q2;
    //fi;
    stVector += LAMP_Q1 * (pc_inc + ((pid-x) << LAMP_X_SHIFT));


    //q2: if
    //::  d_step {y!=255;b[pid] = 0;}  goto q22;
    //:: y==255; goto p;
    //fi
	cond1 = (y==LAMP_MAX_Y);
	cond2 = (y!=LAMP_MAX_Y);
	
    stVector += LAMP_Q2 * ( cond1 * (2L * pc_inc)  + cond2 * (pc_inc - (b << (bit_shift + LAMP_OFFSET_SLOT)) ) );
    
    //q22: if
    //:: y==255; goto NCS;
    //fi;
    stVector += LAMP_Q22 * ( cond1 * (-4L * pc_inc));


    //p: if
    //:: y = pid; goto q3;
    //fi;
    stVector += LAMP_P * (pc_inc + ((pid - y) << LAMP_Y_SHIFT));
    

    //q3: if
    //:: x==pid; goto CS;
    //::  d_step {x!=pid;b[pid] = 0;j = 0;}  goto q4;
    //fi;
	cond1 = (x == pid);
	cond2 = (x != pid);
	
    stVector += LAMP_Q3 * (cond1 * (-5L * pc_inc) + cond2 * (pc_inc - (j << (bit_shift + LAMP_OFFSET_J)) - (b << (bit_shift + LAMP_OFFSET_SLOT))  ) );
        
    //q4: if
    //::  d_step {j<numproc && b[j]==0;j = j+1;}  goto q4;
    //:: j==3; goto q5;
    //fi;
	cond1 = ((j<numproc) && (b_j==0));
	cond2 = (j==numproc);
	
    stVector += LAMP_Q4 * ((cond1 << (bit_shift + LAMP_OFFSET_J))+ cond2 * pc_inc);        //printf("numproc=%lld", numproc);


    //q5: if
    //:: y==pid; goto CS;
    //:: y==255; goto NCS;
    //fi;
	cond1 = (y==pid);
	cond2 = (y==LAMP_MAX_Y);
	
    stVector += LAMP_Q5 * ( cond1 * (-7L * pc_inc) + cond2 * (-8 * pc_inc));
    


    //e1: if
    //:: b[pid] = 0; goto NCS;
    //fi;
    stVector += LAMP_E1 * (-9 * pc_inc - (b << (bit_shift + LAMP_OFFSET_SLOT)));
        

    return stVector;
}



__device__ unsigned long long int execBitLamport2 (long long int pid, long long int numproc, long long int stVector)
{
    
    /* byte b[3];
     byte x=255;
     byte y=255;
     
     
     active proctype P_0() {
     byte j=0;
     
     00: NCS: if
     :: b[pid] = 1; goto q1;
     fi;
     
     01: CS: if
     :: y = 255; goto e1;
     
     fi;
     
     02: q1: if
     :: x = pid; goto q2;
     fi;
     
     03: q2: if
     ::  d_step {y!=255;b[x] = 0;}  goto q22;
     
     :: y==255; goto p;
     
     fi;
     04: q22: if
     :: y==255; goto NCS;
     
     fi;
     
     05: p: if
     :: y = pid; goto q3;
     
     fi;
     
     06: q3: if
     :: x==pid; goto CS;
     
     ::  d_step {x!=pid;b[pid] = 0;j = 0;}  goto q4;
     
     fi;
     
     07: q4: if
     ::  d_step {j<N && b[j]==0;j = j+1;}  goto q4;
     
     :: j==N; goto q5;
     
     fi;
     
     08: q5: if
     :: y==pid; goto CS;
     
     :: y==255; goto NCS;
     
     fi;
     
     09: e1: if
     :: b[pid] = 0; goto NCS;
     
     fi;
     }*/
    
    long long int  bit_shift      = pid * LAMP_STATE_BITSHIFT;
    
    
    //Ezio: I extract the program counter and other important variables
    
    long long int               pc = (stVector & (LAMP_PC_MASK    << bit_shift)) >>  bit_shift;
    long long int                b = (stVector & (LAMP_SLOT_MASK  << bit_shift)) >>  (bit_shift + LAMP_OFFSET_SLOT);
    long long int                j = (stVector & (LAMP_J_MASK     << bit_shift)) >>  (bit_shift + LAMP_OFFSET_J);
    
    long long int          j_shift = j * LAMP_STATE_BITSHIFT;
    
    long long int              b_j =  (stVector & (LAMP_SLOT_MASK  << j_shift))  >>  (j_shift + LAMP_OFFSET_SLOT);
    
    long long int             x    =  (stVector & LAMP_X_MASK) >> LAMP_X_SHIFT;
    long long int             y    =  (stVector & LAMP_Y_MASK) >> LAMP_Y_SHIFT;
    
    long long int          x_shift = x * LAMP_STATE_BITSHIFT;
    
    long long int              b_x =  (stVector & (LAMP_SLOT_MASK  << x_shift)) >>  (x_shift + LAMP_OFFSET_SLOT);
    
    long long int pc_inc           = (1L << bit_shift);
    
	long long int cond1, cond2;
    
    //NCS: if
    //    :: b[pid] = 1; goto q1;
    //fi;
    stVector += LAMP_NCS * (2L * pc_inc + ((1-b) << (bit_shift + LAMP_OFFSET_SLOT)));
    
    //CS: if
    //    :: y = 255; goto e1;
    //fi;
    stVector += LAMP_CS * (8L * pc_inc + ((LAMP_MAX_Y-y) << (LAMP_Y_SHIFT)));

    //q1: if
    //:: x = pid; goto q2;
    //fi;
    stVector += LAMP_Q1 * (pc_inc + ((pid-x) << LAMP_X_SHIFT));

    //q2: if
    //::  d_step {y!=255;b[x] = 0;}  goto q22;
    //:: y==255; goto p;
    //fi
	cond1 = (y==LAMP_MAX_Y);
	cond2 = (y!=LAMP_MAX_Y);
	
    stVector += LAMP_Q2 * ( cond1 * (2L * pc_inc)  + cond2 * (pc_inc - (b_x << (x_shift + LAMP_OFFSET_SLOT)) ) );

    //q22: if
    //:: y==255; goto NCS;
    //fi;
    stVector += LAMP_Q22 * ( (y==LAMP_MAX_Y) * (-4L * pc_inc));
        
    //p: if
    //:: y = pid; goto q3;
    //fi;
    stVector += LAMP_P * (pc_inc + ((pid - y) << LAMP_Y_SHIFT));
        
    //q3: if
    //:: x==pid; goto CS;
    //::  d_step {x!=pid;b[pid] = 0;j = 0;}  goto q4;
    //fi;
	cond1 = (x == pid);
	cond2 = (x != pid);
	
    stVector += LAMP_Q3 * (cond1 * (-5L * pc_inc) + cond2 * (pc_inc - (j << (bit_shift + LAMP_OFFSET_J)) - (b << (bit_shift + LAMP_OFFSET_SLOT))  ) );
    
    //q4: if
    //::  d_step {j<numproc && b[j]==0;j = j+1;}  goto q4;
    //:: j==3; goto q5;
    //fi;
	cond1 = (j<numproc && b_j==0);
	cond2 = (x != pid);
	
    stVector += LAMP_Q4 * (( cond1 << (bit_shift + LAMP_OFFSET_J)) + cond2 * pc_inc);
    
    //q5: if
    //:: y==pid; goto CS;
    //:: y==255; goto NCS;
    //fi;
	cond1 = (y==pid);
	cond2 = (y==LAMP_MAX_Y);
	
    stVector += LAMP_Q5 * (cond1 * (-7L * pc_inc) + cond2 * (-8 * pc_inc));
    
    //e1: if
    //:: b[pid] = 0; goto NCS;
    //fi;
    stVector += LAMP_E1 * (-9 * pc_inc - (b << (bit_shift + LAMP_OFFSET_SLOT)));
    
    return stVector;
}


/*For Lamport 8 */

__device__ unsigned long long int execBitLamport3 (long long int pid, long long int numproc, unsigned long long int stVector)
{
    
    /* byte b[3];
     byte x=255;
     byte y=255;
     
     
     active proctype P_0() {
     byte j=0;
     
     00: NCS: if
     :: b[pid] = 1; goto q1;
     fi;
     
     01: CS: if
     :: y = 255; goto e1;
     
     fi;
     
     02: q1: if
     :: x = pid; goto q2;
     fi;
     
     03: q2: if
     ::  d_step {y!=255;b[pid] = 0;}  goto q22;
     
     :: y==255; goto p;
     
     fi;
     04: q22: if
     :: y==255; goto NCS;
     
     fi;
     
     05: p: if
     :: y = pid; goto q3;
     
     fi;
     
     06: q3: if
     :: x==pid; goto CS;
     
     ::  d_step {x!=pid;b[pid] = 0;j = 0;}  goto q4;
     
     fi;
     
     07: q4: if
     ::  d_step {j<N && b[j]==0;j = j+1;}  goto q4;
     
     :: j==N; goto q5;
     
     fi;
     
     08: q5: if
     :: y==x; goto CS;   --> this is changed
     
     :: y==255; goto NCS;
     
     fi;
     
     09: e1: if
     :: b[pid] = 0; goto NCS;
     
     fi;
     }*/
    
    long long int  bit_shift      = pid * LAMP_STATE_BITSHIFT;
    
    
    //Ezio: I extract the program counter and other important variables
    
    long long int               pc = (stVector & (LAMP_PC_MASK    << bit_shift)) >>  bit_shift;
    long long int                b = (stVector & (LAMP_SLOT_MASK  << bit_shift)) >>  (bit_shift + LAMP_OFFSET_SLOT);
    long long int                j = (stVector & (LAMP_J_MASK     << bit_shift)) >>  (bit_shift + LAMP_OFFSET_J);
    
    long long int          j_shift = j * LAMP_STATE_BITSHIFT;
    
    long long int              b_j =  (stVector & (LAMP_SLOT_MASK  << j_shift))  >>  (j_shift + LAMP_OFFSET_SLOT);
    
    long long int             x    =  (stVector & LAMP_X_MASK) >> LAMP_X_SHIFT;
    long long int             y    =  (stVector & LAMP_Y_MASK) >> LAMP_Y_SHIFT;
    
    long long int          x_shift = x * LAMP_STATE_BITSHIFT;
    
    
    long long int pc_inc           = (1L << bit_shift);
    
    
    
    //NCS: if
    //    :: b[pid] = 1; goto q1;
    //fi;
    stVector += LAMP_NCS * (2L * pc_inc + ((1-b) << (bit_shift + LAMP_OFFSET_SLOT)));
    
    //CS: if
    //    :: y = 255; goto e1;
    //fi;
    stVector += LAMP_CS * (8L * pc_inc + ((LAMP_MAX_Y-y) << (LAMP_Y_SHIFT)));
    
    //q1: if
    //:: x = pid; goto q2;
    //fi;
    stVector += LAMP_Q1 * (pc_inc + ((pid-x) << LAMP_X_SHIFT));
    
    //q2: if
    //::  d_step {y!=255;b[x] = 0;}  goto q22;
    //:: y==255; goto p;
    //fi
    stVector += LAMP_Q2 * ( (y==LAMP_MAX_Y) * (2L * pc_inc)  + (y!=LAMP_MAX_Y) * (pc_inc - (b << (bit_shift + LAMP_OFFSET_SLOT)) ) );
    
    //q22: if
    //:: y==255; goto NCS;
    //fi;
    stVector += LAMP_Q22 * ( (y==LAMP_MAX_Y) * (-4L * pc_inc));
    
    //p: if
    //:: y = pid; goto q3;
    //fi;
    stVector += LAMP_P * (pc_inc + ((pid - y) << LAMP_Y_SHIFT));
    
    //q3: if
    //:: x==pid; goto CS;
    //::  d_step {x!=pid;b[pid] = 0;j = 0;}  goto q4;
    //fi;
    stVector += LAMP_Q3 * ((x == pid) * (-5L * pc_inc) + (x != pid) * (pc_inc - (j << (bit_shift + LAMP_OFFSET_J)) - (b << (bit_shift + LAMP_OFFSET_SLOT))  ) );
    
    //q4: if
    //::  d_step {j<numproc && b[j]==0;j = j+1;}  goto q4;
    //:: j==3; goto q5;
    //fi;
    stVector += LAMP_Q4 * (((b_j==0) * (long long int)(j<numproc) << (bit_shift + LAMP_OFFSET_J))+ (j==numproc) * pc_inc);
    
    //q5: if
    //:: y==x; goto CS;       --> Changed
    //:: y==255; goto NCS;
    //fi;
    stVector += LAMP_Q5 * ((y==x) * (-7L * pc_inc) + (y==LAMP_MAX_Y) * (-8 * pc_inc));
    
    //e1: if
    //:: b[pid] = 0; goto NCS;
    //fi;
    stVector += LAMP_E1 * (-9 * pc_inc - (b << (bit_shift + LAMP_OFFSET_SLOT)));
    
    return stVector;
}













/*__global__ void ezio_test(unsigned long long int * test)        
{
    unsigned long long int temp = test[0];
    test[0] = execBit(1,3,temp); 
}*/





unsigned long long int * createQueuesSlots ()
{
    unsigned long long int * queues_slots =  (unsigned long long int *) malloc (2 * NxNxMAX_QUEUE_NUM_SLOTS * sizeof(unsigned long long int));
    for (int i=0; i < 2 * NxNxMAX_QUEUE_NUM_SLOTS; i++)
    {
        queues_slots[i] = 18446744073709551615;//because unsigned
    } 
    return queues_slots;
}

Queue * createCoresQueues ()
{
    Queue * cores_queues = (Queue *) malloc (2 * NxN * sizeof(Queue));
    Queue * temp;

    /*Inizialization*/

    for (int t=0; t < 2; t++)
    {
        for (int w=0; w < N; w++)
        {
            for (int q=0; q < N; q++)
            {
                 temp       = cores_queues + t*NxN + w*N + q;
                 temp->start = t*NxNxMAX_QUEUE_NUM_SLOTS + w*NxMAX_QUEUE_NUM_SLOTS + q*MAX_QUEUE_NUM_SLOTS; 
                 temp->stop  = temp->start;                               
            }
        }
    }
    return cores_queues;
}

Functions * createFunctions ()//new
{
      Functions * function_set = (Functions *) malloc (sizeof(Functions));
      //function_set->Generate();
      return function_set;

}


Functions * createGPUFunctions ()//new
{
      Functions * gpu_function_set;
      cudaMalloc( (void**)&gpu_function_set, sizeof(Functions));
      //gpu_function_set->Generate();
      return gpu_function_set;

}


void destroyFunctions(Functions * function_set)//new
{
      free(function_set);
}



Queue * createGPUCoresQueues ()
{
      Queue * gpu_cores_queues;
      cudaMalloc( (void**)&gpu_cores_queues, 2 * NxNxMAX_QUEUE_NUM_SLOTS * sizeof(Queue));
      return gpu_cores_queues;
}

unsigned long long int * createGPUQueuesSlots ()
{
      unsigned long long int * gpu_queues_slots;
      cudaMalloc( (void**)&gpu_queues_slots, 2 * NxNxMAX_QUEUE_NUM_SLOTS * sizeof(unsigned long long int));
      return gpu_queues_slots;
}

int * createGPUVector (int num_elements)
{
      int * gpu_vector;
      cudaMalloc ( (void**)&gpu_vector, num_elements * sizeof(int));
      return gpu_vector;
}

void copy_vectors_CPU2GPU (int * cpu_vector, int * gpu_vector, int num_elements)
{
     cudaMemcpy (gpu_vector, cpu_vector, num_elements * sizeof(int), cudaMemcpyHostToDevice);
}

void copy_vectors_GPU2CPU (int * gpu_vector, int * cpu_vector, int num_elements)
{
     cudaMemcpy (cpu_vector, gpu_vector, num_elements * sizeof(int), cudaMemcpyDeviceToHost);
}


void copy_Constant_CPU2GPU(int * cpuCFG)
{
	cudaMemcpyToSymbol(consCFG, cpuCFG, sizeof(int)*16/*, 0, cudaMemcpyHostToDevice*/);
}

void copy_Constant_CPU2GPU_2D(int * cpuOrder)
{
	cudaMemcpyToSymbol(consOrder, cpuOrder, sizeof(int)*96/*, 0, cudaMemcpyHostToDevice*/);
}

void copy_Constant_CPU2GPU_Rep(unsigned long long int * cpuRnd)
{
	cudaMemcpyToSymbol(consRnd, cpuRnd, sizeof(unsigned long long int)*100/*, 0, cudaMemcpyHostToDevice*/);
}



unsigned long long int * createLLGPUVector (int num_elements)
{
      unsigned long long int * gpu_vector;
      cudaMalloc ( (void**)&gpu_vector, num_elements * sizeof(unsigned long long int));
      return gpu_vector;
}

void copy_llvectors_CPU2GPU (unsigned long long int * cpu_vector, unsigned long long int * gpu_vector, int num_elements)
{
     cudaMemcpy (gpu_vector, cpu_vector, num_elements * sizeof(unsigned long long int), cudaMemcpyHostToDevice);
}

void copy_llvectors_GPU2CPU (unsigned long long int * gpu_vector, unsigned long long int * cpu_vector, int num_elements)
{
     cudaMemcpy (cpu_vector, gpu_vector, num_elements * sizeof(unsigned long long int), cudaMemcpyDeviceToHost);
}








//new
uint8_t * create8bGPUVector (int num_elements)
{
      uint8_t * gpu_vector;
      cudaMalloc ( (void**)&gpu_vector, num_elements * sizeof(uint8_t));
      return gpu_vector;
}

//new
void copy_8bvectors_GPU2CPU (uint8_t * gpu_vector, uint8_t * cpu_vector, int num_elements)
{
     cudaMemcpy (cpu_vector, gpu_vector, num_elements * sizeof(uint8_t), cudaMemcpyDeviceToHost);
}









//new
bool * createBGPUVector (int num_elements)
{
      bool * gpu_vector;
      cudaMalloc ( (void**)&gpu_vector, num_elements * sizeof(bool));
      return gpu_vector;
}

//new
void copy_Bvectors_GPU2CPU (bool * gpu_vector, bool * cpu_vector, int num_elements)
{
     cudaMemcpy (cpu_vector, gpu_vector, num_elements * sizeof(bool), cudaMemcpyDeviceToHost);
}



void copyQUEUES_GPU2CPU (Queue * gpu_cores_queues, Queue * cores_queues, unsigned long long int * gpu_queues_slots, unsigned long long int * queues_slots)
{
     cudaMemcpy (cores_queues, gpu_cores_queues, 2 * NxN * sizeof(Queue),                      cudaMemcpyDeviceToHost);
     cudaMemcpy (queues_slots, gpu_queues_slots, 2 * NxNxMAX_QUEUE_NUM_SLOTS * sizeof(unsigned long long int),    cudaMemcpyDeviceToHost);
}


void copyReportedStates_GPU2CPU (int * reported_states, int * gpu_reported_states, int n)
{
     cudaMemcpy (reported_states, gpu_reported_states, n * sizeof(int), cudaMemcpyDeviceToHost);
}



void copyQUEUES_CPU2GPU (Queue * gpu_cores_queues, Queue * cores_queues, unsigned long long int * gpu_queues_slots, unsigned long long int * queues_slots)
{
     cudaMemcpy (gpu_cores_queues, cores_queues, 2 * NxN * sizeof(Queue),                      cudaMemcpyHostToDevice);
     cudaMemcpy (gpu_queues_slots, queues_slots, 2 * NxNxMAX_QUEUE_NUM_SLOTS * sizeof(unsigned long long int),    cudaMemcpyHostToDevice);     
}

void setInitialState (Queue * cores_queues, unsigned long long int * queues_slots, unsigned long long int state)
{
     queues_slots[CQUEUES(0,0,0).stop] = state;
     CQUEUES(0,0,0).stop++;
     queues_slots[CQUEUES(0,0,0).stop] = 18446744073709551615;//because unsigned
}


void destroyQueuesSlots(unsigned long long int * queues_slots)
{
     free(queues_slots);
}

void destroyCoresQueues (Queue * cores_queues)
{
     free(cores_queues);
}

/*void printQueue (Queue * cores_queues, int * queues_slots, int t, int w, int q)
{
     int num_elements = CQUEUES(t,w,q).stop - CQUEUES(t,w,q).start;
     printf("Start:%d Stop:%d\n",CQUEUES(t,w,q).start, CQUEUES(t,w,q).stop);     
     printf("Queue Q(%d,%d,%d) = [", t, w, q);
     
     for (int i=0; i < num_elements; i++)
     {
          printf("%d ",queues_slots[CQUEUES(t,w,q).start + i]);    
     }

     printf("]\n num_elements=%d\n",num_elements); 
}*/



