#include <vector_types.h>
#include <stdio.h>
#include <ctime>
#include <string.h>
#include <math.h>
#include <curand.h>
#include <cuda.h>
#include <curand_kernel.h>
#include <cuda_runtime.h>

//25 works, 20 works, 3 works, 10 works, 9 works
#define N                              32//32
#define MAX_QUEUE_NUM_SLOTS        4//3 4
//may need to drastically cut slots if we want the queues and the hash in shared memory
//if 25, global memory.  nxslots=800, nxnxslots=25600, nxnxslotsx64= 1638400 (more than 48KB shared memory)
//if 5: 327,680 bits for 5120 queued (64-bit) states, leaving 69053 bits for hashing
//if 3: 196,608 bits for 3072 queued (64-bit) states, leaving 196608 bits for hashing
#define NxN                           1024//1024
#define NxMAX_QUEUE_NUM_SLOTS     128//96 128
#define NxNxMAX_QUEUE_NUM_SLOTS  4096//3072 4096

#define NUMHASHFUNC     1//number of hash functions
//393216 bits of shared memory
//prime larger than that is 396733
#define kPrimeDivisor 396733u //larger than the largest hash table possible in shared memory
//#define kPrimeDivisor   4294967291u//theoretically larger than the largest practical hash table size

#define NUM_VERTEX  15
#define QUEUES_SPACE    N*N*NUM_VERTEX
#define Q(t,w,q)        Q[t*QUEUES_SPACE + w*N*NUM_VERTEX + q*NUM_VERTEX]
#define QS(t,w,q,s)     Q[t*QUEUES_SPACE + w*N*NUM_VERTEX + q*NUM_VERTEX + s]


#define CQUEUES(t,w,q)        cores_queues[t*NxN + w*N + q]
#define CQUEUE(t,w,q)         queues_slots[t*NxNxMAX_QUEUE_NUM_SLOTS + w*NxMAX_QUEUE_NUM_SLOTS + q*MAX_QUEUE_NUM_SLOTS]
#define CQUEUE_SLOT(t,w,q,s)  queues_slots[t*NxNxMAX_QUEUE_NUM_SLOTS + w*NxMAX_QUEUE_NUM_SLOTS + q*MAX_QUEUE_NUM_SLOTS + s]

#define CUDA_CALL(x) do { if((x)!=cudaSuccess) { \
printf("Error at %s:%d\n",__FILE__ ,__LINE__);\
return EXIT_FAILURE ;}} while (0)

#define CURAND_CALL(x) do { if((x)!=CURAND_STATUS_SUCCESS) { \
printf("Error at %s:%d\n",__FILE__ ,__LINE__);\
return EXIT_FAILURE ;}} while (0)




//our main data structure
struct Queue {
       unsigned long long int start;
       unsigned long long int stop;
};



void GenerateFunctions(uint2 *constants);




//! Container for all of the hash functions.
struct Functions {
  //! The constants required for all of the hash functions, including the stash.  Each function requires 2.
  uint2 constants[NUMHASHFUNC+1];
  void Generate() {GenerateFunctions(constants);}
};




__device__ bool checkRnd(unsigned long long int state);









__global__ void stateSpaceExplorationFPGAPDS (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states,
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a,
				unsigned long long int mix_b);



__global__ void stateSpaceExplorationFPGAEven (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states,
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a,
				unsigned long long int mix_b);





__global__ void stateSpaceExplorationFPGAOdd (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states,
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a,
				unsigned long long int mix_b);





__global__ void stateSpaceExplorationFPGA (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states,
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a,
				unsigned long long int mix_b);







__global__ void stateSpaceExplorationAB32BitOdd (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states,
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a,
				unsigned long long int mix_b);


__global__ void stateSpaceExplorationAB32BitEven (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states,
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a,
				unsigned long long int mix_b);




__global__ void stateSpaceExplorationAB32BitNDInR (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states,
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a,
				unsigned long long int mix_b);

__global__ void stateSpaceExplorationAB32BitNDOutR (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states,
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a,
				unsigned long long int mix_b);





__global__ void stateSpaceExplorationAB32BitNDIn (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states,
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a,
				unsigned long long int mix_b);

__global__ void stateSpaceExplorationAB32BitNDOut (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states,
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a,
				unsigned long long int mix_b);







__global__ void stateSpaceExplorationAB32Bit (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states,
				int numstates,
                        curandState * cstate,
				uint8_t * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a,
				unsigned long long int mix_b);





__global__ void stateSpaceExplorationAB32S (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states,
				int numstates,
                        curandState * cstate,
				bool * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a,
				unsigned long long int mix_b);

__global__ void stateSpaceExplorationAB32 (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states,
				int numstates,
                        curandState * cstate,
				bool * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a,
				unsigned long long int mix_b);



__global__ void stateSpaceExplorationAB (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states,
				int numstates,
                        curandState * cstate,
				bool * table,
				int table_size, unsigned long long int * rep_table,
				unsigned long long int mix_a,
				unsigned long long int mix_b);


__global__ void stateSpaceExploration (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states,
				int numstates,
                        curandState * cstate,
				bool * table,
				int table_size);
//reporting
__global__ void stateSpaceExploration (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states,
				int numstates,
                        curandState * cstate,
				bool * table,
				int table_size, unsigned long long int * rep_table);



__global__ void stateSpaceExplorationND (Queue * cores_queues,
				Functions * function_set, 
                                unsigned long long int * queues_slots,
                                bool * reported_states,
				int numstates,
                        curandState * cstate,
				bool * table,
				int table_size);




__device__ void gpu_lock_free_sync(int goalVal, volatile int * Arrayin, volatile int * Arrayout);

__device__ unsigned long long int execBit(int iproc, int numproc, unsigned long long int stVector);
__device__ unsigned long long int execBitAE(int iproc, int numproc, unsigned long long int stVector);
__device__ unsigned long long int execBitAO(int iproc, int numproc, unsigned long long int stVector);

__device__ unsigned long long int execBitGen32 (long long int pid, int nd, unsigned long long int stVector);

__device__ unsigned long long int execBitPeterson1(long long int pid, long long int nd, long long int numproc, unsigned long long int stVector);
__device__ unsigned long long int execBitPeterson2(long long int pid, long long int nd, long long int numproc, unsigned long long int stVector);
__device__ unsigned long long int execBitPeterson3(long long int pid, long long int nd, long long int numproc, unsigned long long int stVector);

__device__ unsigned long long int execBitLamport1(long long int pid, long long int numproc, unsigned long long int stVector);
__device__ unsigned long long int execBitLamport2(long long int pid, long long int numproc, long long int stVector);
__device__ unsigned long long int execBitLamport3(long long int pid, long long int numproc, unsigned long long int stVector);

Functions * createFunctions();
Functions * createGPUFunctions();
void destroyFunctions(Functions * function_set);

unsigned long long int * createQueuesSlots ();
Queue * createCoresQueues ();
Queue * createGPUCoresQueues ();
Queue * createCoresQueues ();
Queue * createGPUCoresQueues ();
unsigned long long int * createGPUQueuesSlots ();

int * createGPUVector (int num_elements);
void copy_vectors_CPU2GPU (int * cpu_vector, int * gpu_vector, int num_elements);
void copy_vectors_GPU2CPU (int * gpu_vector, int * cpu_vector, int num_elements);

void copy_Constant_CPU2GPU(int * cpuCFG);
void copy_Constant_CPU2GPU_2D(int * cpuorder);
void copy_Constant_CPU2GPU_Rep(unsigned long long int * cpuRnd);

unsigned long long int * createLLGPUVector (int num_elements);
void copy_llvectors_CPU2GPU (unsigned long long int * cpu_vector, unsigned long long int * gpu_vector, int num_elements);
void copy_llvectors_GPU2CPU (unsigned long long int * gpu_vector, unsigned long long int * cpu_vector, int num_elements);


uint8_t * create8bGPUVector (int num_elements);
void copy_8bvectors_GPU2CPU (uint8_t * gpu_vector, uint8_t * cpu_vector, int num_elements);


bool * createBGPUVector (int num_elements);
void copy_Bvectors_GPU2CPU (bool * gpu_vector, bool * cpu_vector, int num_elements);

void copyQUEUES_GPU2CPU (Queue * gpu_cores_queues, Queue * cores_queues, unsigned long long int * gpu_queues_slots, unsigned long long int * queues_slots);
void copyReportedStates_GPU2CPU (int * reported_states, int * gpu_reported_states, int n);

void copyQUEUES_CPU2GPU (Queue * gpu_cores_queues, Queue * cores_queues, unsigned long long int * gpu_queues_slots, unsigned long long int * queues_slots);

void setInitialState (Queue * cores_queues, unsigned long long int * queues_slots, unsigned long long int state);
void destroyQueuesSlots(int * queues_slots);
void destroyCoresQueues (Queue * cores_queues);
