import sys
import matplotlib
import matplotlib.pyplot as plt
import csv
import numpy as np
import matplotlib.font_manager
import matplotlib.colors

matplotlib.font_manager._rebuild()

input_file = ""

#=======================
# read CSV file
#=======================
def read_csv(input_file):

	with open(input_file, "r", encoding="utf8") as source:
		csv_reader = list(csv.reader(open(input_file, "r"), delimiter=","))

	headerRowToInclude = 1
	unit = 1
	skippedrows = 2
	skippedendrows = 1
	skippedcolumns = 8
	configs_size = 10
	configs = []

	header=csv_reader[headerRowToInclude]
	del header[0:skippedcolumns]

	nCols = skippedcolumns + configs_size
	nRows = len(csv_reader) - skippedendrows
	for i in range(skippedcolumns,nCols): 
		model_result = []
		for j in range(skippedrows, nRows + skippedrows):
			if 'Average' in csv_reader[j][0]: 
				break
			speed = csv_reader[j][i]
			if '*' in speed: 
				speed = speed.replace('*','')
			if '---' in speed: 
				speed = speed.replace('---','0')
			speed = float(speed) / unit
			model_result.append(speed)
		configs.append(model_result)

	for config in configs:
		config.sort()

	return header, configs

#======================
# plot
#======================
def plot_csv(filename, header, configs, CDF = 1): 
	configs_size = len(configs)
	fontSize = 40
	mS = 12
	mS2 = 15
	mW = 2
	lW = 2.8
	xstart = 0
	ystart = 0
	modelsstep = 2 if len(configs[0]) > 10 else 1
	modelslimit = len(configs[0])
	speedlimitstep = 10
	speedlimit = 0
	for config in configs:
		maxPerConfig = max(config)
		if speedlimit < maxPerConfig:
			speedlimit = maxPerConfig
	fillmarker = 'full'

	xvals = []

	for i in range(0, configs_size):
		xvals.append( (header[i] , range(xstart, len(configs[i]))) )

	markers    = ['h','>','x','s','+','o','*','p','<', '^']
	markersize = [mS2, mS, mS2, mS, mS2, mS, mS2, mS2, mS, mS]
	colors = ['tab:gray','cadetblue','tab:blue','tab:orange','slateblue','tab:purple','tab:green', 'sandybrown', 'tab:olive', 'tab:red','cadetblue','darkviolet','coral','goldenrod']

	fig = plt.figure(figsize=(25,13))
	plt.rcParams["font.family"] = "Times New Roman"

	speedlabel = 'Millions of states per seconds'
	modelslabel = 'SLCO Models'

	for i in range(0, configs_size):
		if CDF:
			xLab = speedlabel
			yLab = modelslabel
			plt.plot(configs[i], xvals[i][1], linestyle='-', marker=markers[i], markeredgewidth=mW, markersize=markersize[i], fillstyle=fillmarker, linewidth=lW, color=colors[i])
		else:
			yLab = speedlabel
			xLab = modelslabel
			plt.plot(xvals[i][1], configs[i], linestyle='-', marker=markers[i], markeredgewidth=mW, markersize=markersize[i], fillstyle=fillmarker, linewidth=lW, color=colors[i])

	plt.legend(header, fontsize=fontSize, loc="best")

	if CDF:
		plt.yticks(np.arange(xstart, modelslimit + 1, modelsstep), fontsize=fontSize)
		plt.xticks(np.arange(ystart, speedlimit + 1, speedlimitstep), fontsize=fontSize)
	else:
		plt.yticks(np.arange(ystart, speedlimit + 1, speedlimitstep), fontsize=fontSize)
		plt.xticks(np.arange(xstart, modelslimit + 1, modelsstep), fontsize=fontSize)

	ax = plt.gca()
	ax.axes.yaxis.get_offset_text().set_fontsize(fontSize)
	ax.get_yaxis().set_major_formatter(matplotlib.ticker.FuncFormatter(lambda x, p: format(int(x), ',')))
	#ax.annotate('Timeout', (timeout + 20, 145), color='black', size=20)

	if CDF:
		plt.xlim(left=-1)
		plt.ylim(bottom=-1)
		ax.yaxis.get_major_ticks()[0].set_visible(False)
	else:
		plt.xlim(left=0)
		plt.ylim(bottom=0)

	ax.tick_params(axis="y",direction="in", length=8, right=True)
	ax.tick_params(axis="x",direction="in", length=8, top=True)
	plt.xlabel(xLab, fontsize=fontSize, fontweight='normal', labelpad=25)
	plt.ylabel(yLab, fontsize=fontSize, fontweight='normal', labelpad=25)
	#plt.grid(linestyle='dotted')
	plt.tight_layout(pad=2)
	plt.savefig(filename + '.pdf',dpi=1000)
	#plt.show()

def main(args):
	global input_file

	if len(args) == 0:
		print("Missing argument: CSV file.")
		sys.exit(1)
	else:
		if args[0] == '-h' or args[0] == '-help':
			print("Usage: python3 genfigure.py <csv file>")
			print("")
			print(" -f   parse CSV file.")
			sys.exit(0)
		else:
			for i in range(0,len(args)):
				if args[i] == '-f':
					input_file = args[i+1]
					i += 1

	header, configs = read_csv(input_file)
	plot_csv(input_file.replace('.csv', ''), header, configs)

if __name__ == '__main__':
	args = []
	for i in range(1,len(sys.argv)):
		args.append(sys.argv[i])
	main(args)