#!/bin/bash

export LC_NUMERIC="en_US.UTF-8"

models_dir=$1
resultsdir=$2
nruns=$3

[ -z $models_dir ] && echo "models directory not given" && exit 
[ -z $resultsdir ] && echo "artifact results directory not given" && exit 
[ -z $nruns ] && echo "number of iterations not given" && exit 

for model in $models_dir/*.gpf 
do
	modelname=${model##*/} 
	modelname=${modelname%.*} 

	resultlogs=$resultsdir/gpuexplore2/$modelname

	mkdir -p $resultlogs

	time=0
	states=0
	avgSpeed=0
	memoutsig=0

	for ((i=0;i<$nruns;i++))
	do
		ntime=$(( $i + 1 ))
		printf " Running GPUexplore 2.0 on ( %-11s ) the %s-th time... \r" "$modelname" "$ntime"
		
		logfile=$resultlogs/log_run${i}.txt
		
		./gpuexplore2 ${model%.*} > $logfile 2>&1
		
		line=$(grep 'ERROR' $logfile)
		[ ! -z "$line" ] && memoutsig=1
		
		if [ $i = 0 ]; then 	
			line=$(grep 'nr. of states in hash table' $logfile)
			[ ! -z "$line" ] && states=$(echo $line | awk '{ print $7 }')
		fi
		
		line=$(grep 'Exploration time' $logfile)
	
		if [ ! -z "$line" ]; then 
			time=$(echo $line | awk '{ print $3 }')
			speed=$(echo "scale=0;$states / $time" | bc -l)
			avgSpeed=$(echo "$avgSpeed + $speed" | bc -l)
		fi
		
		[ $nruns -gt 1 ] && [ $memoutsig = 1 ] && break
	done

	mainresult=$resultlogs/average_all.txt

	echo -n "" > $mainresult

	tailstr="DONE"
	[ $memoutsig = 1 ] && tailstr="MEMOUT"
	
	avgSpeed=$(echo "scale=0;$avgSpeed / $nruns" | bc -l)

	echo $tailstr >> $mainresult

	printf "Runs                     : %-10d\n" $ntime >> $mainresult
	printf "States                   : %-10d\n" $states >> $mainresult
	printf "States per second (mean) : %-10d\n" $avgSpeed >> $mainresult

done 

nModels=$(ls $models_dir | wc -l)

printf " Running GPUexplore 2.0  on ( $nModels %-11s ) for %s times... DONE.\n" "other models" "$nruns"
