#!/bin/bash

export LC_NUMERIC="en_US.UTF-8"

modelpath=$1
configopt=$2 # if empty the geenrator resorts to the default settings
resultdir=$3
nruns=$4

[ ! -f $modelpath ] && echo "error: model '$modelpath' does not exist" && exit
[ ! -d $resultdir ] && echo "error: artifact results '$resultdir' not found" && exit

sortedConfigopt=$(echo $configopt | xargs -n1 | sort -u | xargs)

COMBO1="cmp cu i1"
COMBO2="cmp bu i1"
COMBO3="cmp cu bu i1"
COMBO4="bu i30"
COMBO5="cmp i30"
COMBO6="cmp cu i30"
COMBO7="cmp bu i30"
COMBO8="cmp cu bu i30"

COMBO1=$(echo $COMBO1 | xargs -n1 | sort -u | xargs)
COMBO2=$(echo $COMBO2 | xargs -n1 | sort -u | xargs)
COMBO3=$(echo $COMBO3 | xargs -n1 | sort -u | xargs)
COMBO4=$(echo $COMBO4 | xargs -n1 | sort -u | xargs)
COMBO5=$(echo $COMBO5 | xargs -n1 | sort -u | xargs)
COMBO6=$(echo $COMBO6 | xargs -n1 | sort -u | xargs)
COMBO7=$(echo $COMBO7 | xargs -n1 | sort -u | xargs)
COMBO8=$(echo $COMBO8 | xargs -n1 | sort -u | xargs)

# '-l' option skips the GPU quering in the code generator
generator_config=""
gpuexplore_config=""
if [[ "$sortedConfigopt" == "bu i1" ]]; then 
	generator_config="-l -nocuckoo -norootcompression -nosmartfetching"
elif [[ "$sortedConfigopt" == "cmp i1" ]]; then 
	generator_config="-l -nocuckoo -nobuckets -nosmartfetching"
elif [[ "$sortedConfigopt" == "$COMBO1" ]]; then 
	generator_config="-l -nobuckets -nosmartfetching"
elif [[ "$sortedConfigopt" == "$COMBO2" ]]; then 
	generator_config="-l -nocuckoo -nosmartfetching"
elif [[ "$sortedConfigopt" == "$COMBO3" ]]; then 
	generator_config="-l -nosmartfetching"; 
elif [[ "$sortedConfigopt" == "$COMBO4" ]]; then 
	generator_config="l -nocuckoo -norootcompression -nosmartfetching"
	gpuexplore_config="-k 30"
elif [[ "$sortedConfigopt" == "$COMBO5" ]]; then 
	generator_config="-l -nocuckoo -nobuckets -nosmartfetching"
	gpuexplore_config="-k 30"
elif [[ "$sortedConfigopt" == "$COMBO6" ]]; then 
	generator_config="-l -nobuckets -nosmartfetching"; 
	gpuexplore_config="-k 30"
elif [[ "$sortedConfigopt" == "$COMBO7" ]]; then 
	generator_config="-l -nocuckoo -nosmartfetching"; 
	gpuexplore_config="-k 30"
elif [[ "$sortedConfigopt" == "$COMBO8" ]]; then 
	generator_config="-l -nosmartfetching"; 
	gpuexplore_config="-k 30"
else 
	echo "incorrect combination is given"; exit
fi

configopt=${configopt//" "/"_"}

[ -z "$generator_config" ] && echo "GPU configuration not given" && exit

modelname=${modelpath##*/} 
modelname=${modelname%.*} 

modelbuild=generated/$modelname/$configopt
toolname=gpuexplore3
resultdest=$resultdir/$toolname/$modelname/$configopt

mkdir -p $modelbuild $resultdest

mainresult=$resultdest/average_all.txt

# generate GPUExplore binary
#echo -n " Generating GPUExplore for $modelname with \"$configopt\" configuration..."

logfile=$modelbuild/log_gen_and_make.txt
rm -rf $modelbuild/*.cuh $modelbuild/*.cu $modelbuild/$toolname $modelbuild/$modelname
python3 slco2gpuexplore.py $generator_config -o $modelbuild $modelpath > $logfile 2>&1
gpuexpheader=$modelbuild/${modelname}_gpuexplore.cuh
gpuexpmain=$modelbuild/gpuexplore.cu
[ ! -f $gpuexpheader ] && echo "GPUExplore header file not generated" && exit
[ ! -f $gpuexpmain ] && echo "GPUExplore main file not generated" && exit
line=$(grep 'System state vector size' $logfile)
nbits=$(echo $line | awk '{ print $5 }')
line=$(grep 'Number of vector parts' $logfile)
nparts=$(echo $line | awk '{ print $5 }')
make -C $modelbuild >> $logfile 2>&1
[ ! -f $modelbuild/gpuexplore ] && echo "GPUExplore binary for $modelname not generated" && exit
chmod +x $modelbuild/gpuexplore
mv $modelbuild/gpuexplore $modelbuild/$modelname

#echo -n " done."

# testing starts from here
timelimit=3600
time=0
nodes=0
states=0
avgSpeed=0
memoutsig=0
timeoutsig=0

chmod +x ../resourcelimit

[ "$modelname" = "elevator2.5" ] && nruns=1

for (( iter=0; iter<$nruns; iter++ ))
do
	ntime=$(( $iter + 1 ))
	configtoprint=${configopt//"_"/" + "}
	printf " Running GPUExplore 3.0 on ( %-11s ) with ( %-8s ) configuration the %s-th time... \r" "$modelname" "$configtoprint" "$ntime"
	
	logfile=$resultdest/log_run${ntime}.txt

	../resourcelimit -t $timelimit $modelbuild/$modelname $gpuexplore_config > $logfile 2>&1

	line=$(grep 'TIMEOUT' $logfile)
	[ ! -z "$line" ] && timeoutsig=1 && break
	
	line=$(grep 'ERROR' $logfile)
	[ ! -z "$line" ] && memoutsig=1
	
	if [ $iter = 0 ]; then 	
		line=$(grep 'Total reachable states' $logfile)
		[ ! -z "$line" ] && states=$(echo $line | awk '{ print $4 }') && states=${states/","/}
		line=$(grep 'Total internal nodes' $logfile)
		[ ! -z "$line" ] && nodes=$(echo $line | awk '{ print $4 }')
	fi
	
	line=$(grep 'Exploration time' $logfile)
	
	if [ ! -z "$line" ]; then 
		time=$(echo $line | awk '{ print $3 }')
		speed=$(echo "scale=0;$states / $time" | bc -l)
		avgSpeed=$(echo "$avgSpeed + $speed" | bc -l)
	fi
	
	[ $nruns -gt 1 ] && [ $memoutsig = 1 ] && break
	[ $nruns -gt 1 ] && [ $timeoutsig = 1 ] && break
done

echo -n "" > $mainresult

tailstr="DONE"
[ $memoutsig = 1 ] && tailstr="MEMOUT"
[ $timeoutsig = 1 ] && tailstr="TIMEOUT"
if [ $timeoutsig = 0 ]; then
	avgSpeed=$(echo "scale=0;$avgSpeed / $nruns" | bc -l)
	if (( $(echo "$nodes > 0" |bc -l) )); then 	
		ratio=$(echo "scale=2;($states * $nparts) / ($states + $nodes)" | bc -l)
	fi
fi
echo $tailstr >> $mainresult

printf "Runs                     : %-10d\n" $ntime >> $mainresult
printf "Bits                     : %-10d\n" $nbits >> $mainresult
printf "States                   : %-10d\n" $states >> $mainresult
printf "States per second (mean) : %-10d\n" $avgSpeed >> $mainresult
printf "Compression ratio        : %-2.2f\n" $ratio >> $mainresult

