#!/bin/bash

export LC_NUMERIC="en_US.UTF-8"

models_dir=$1
resultsdir=$2
niters=$3

[ -z $models_dir ] && echo "models directory not given" && exit 
[ -z $resultsdir ] && echo "artifact results directory not given" && exit 
[ -z $niters ] && echo "number of iterations not given" && exit 

for model_dir in $models_dir/*
do
	rm -rf $model_dir/OUT
	make -C $model_dir clean > /dev/null 2>&1  
	make -C $model_dir > /dev/null 2>&1  
	
	modelname=${model_dir##*/} 
	resultlogs=$resultsdir/grapple/$modelname
	mkdir -p $resultlogs
	avgTime=0
	avgStates=0
	avgSpeed=0
	maxStates=0
	maxTime=0
	for ((i=0;i<$niters;i++))
	do
		ntime=$(( $i + 1 ))
		printf " Running Grapple on ( %-11s ) the %s-th time... \r" "$modelname" "$ntime"
		
		resultfile=$resultlogs/log_run${i}.txt
		
		$model_dir/hashfreeSpin > $resultfile 2>&1
		
		line=$(sed -n '1p' < $resultfile)
		time=$(echo $line | grep -Eo '[0-9]+([.][0-9]+)?')
		
		line=$(sed -n '3p' < $resultfile)
		states=$(echo $line | grep -Eo '[0-9]+')
		
		speed=$(echo "scale=3;$states / $time" | bc -l)
		
		avgTime=$(echo "$avgTime + $time" | bc -l)
		avgStates=$(echo "$avgStates + $states" | bc -l)
		avgSpeed=$(echo "$avgSpeed + $speed" | bc -l)
		
		if (( $(echo "$maxStates < $states" |bc -l) )); then
			maxStates=$states
			maxTime=$time
		fi
		
		#printf " %d: Speed = %8d / %-8.3f = %-8.3f  states/second\n" $i $states $time $speed
	done

	avgStates=$(echo "scale=0;$avgStates / $niters" | bc -l)
	avgSpeed=$(echo "scale=0;$avgSpeed / $niters" | bc -l)
	avgTime=$(echo "scale=3;$avgTime / $niters" | bc -l)

	mainresult=$resultlogs/average_all.txt

	echo -n "" > $mainresult
	printf "Max running time         : %-8.2f  seconds\n" $maxTime >> $mainresult
	printf "Max reachable states     : %-10d   states\n" $maxStates >> $mainresult
	printf "Average time             : %-8.2f  second/run\n" $avgTime >> $mainresult
	printf "Average states           : %-10d\n" $avgStates >> $mainresult
	printf "States per second (mean) : %-10d\n" $avgSpeed >> $mainresult

	[ -f OUT ] && sort -u OUT > $model_dir/OUT && rm -rf OUT 
done 

nModels=$(ls $models_dir | wc -l)

printf " Running Grapple on ( $nModels %-11s ) for %s times... DONE.\n" "other models" "$niters"
