#!/bin/bash
#===================================
# Muhammad Osama, TACAS'23 Artifact
#===================================

export LC_NUMERIC="en_US.UTF-8"

modelpath=$1
ncores=$2
resultdir=$3
nruns=$4

[ ! -f $modelpath ] && echo "error: model '$modelpath' does not exist" && exit
[ ! -d $resultdir ] && echo "error: artifact results '$resultdir' not found" && exit

DEP=divine
[ ! -f $DEP ] && echo "LTSmin dependeny not found" && exit
BINDIR=~/bin
mkdir -p $BINDIR
if [ ! -f $BINDIR/$DEP ]; then 
	cp $DEP $BINDIR
fi

LTSMINBIN=dve2lts-mc
[ ! -f $LTSMINBIN ] && echo "LTSmin binary not found" && exit

chmod +x $DEP $LTSMINBIN

modelname=${modelpath##*/} 
modelname=${modelname%.*} 

resultdest=$resultdir/ltsmin/$modelname/$ncores

mkdir -p $resultdest

mainresult=$resultdest/average_all.txt

time=0
states=0
avgSpeed=0
timelimit=3600  
memlimit=31 # bits
memoutsig=0
timeoutsig=0

[ "$modelname" = "elevator2.5" ] && nruns=1

for (( iter=0; iter<$nruns; iter++ ))
do
	ntime=$(( $iter + 1 ))
	
	printf " Running LTSmin on ( %-11s ) with ( %d-core ) configuration the %s-th time... \r" $modelname "$ncores" "$ntime"

	logfile=$resultdest/log_run${ntime}.txt
	
	./$LTSMINBIN --timeout=$timelimit --size=$memlimit --threads=$ncores $modelpath > $logfile 2>&1
				
	if [ $iter = 0 ]; then
		line=$(grep 'Explored' $logfile)
		if [ ! -z "$line" ]; then
			[ $ncores -gt 1 ] && line=${line/"( 0/ $ncores)"/}
			states=$(echo $line | awk '{ print $3 }')
		fi
	fi
	
	line=$(grep 'States per second' $logfile)
	if [ ! -z "$line" ]; then 
		[ $ncores -gt 1 ] && line=${line/"( 0/ $ncores)"/}
		speed=$(echo $line | awk '{ print $5 }')
		speed=${speed/","/}
		avgSpeed=$(echo "$avgSpeed + $speed" | bc -l)
	fi 
	
	line=$(grep 'TIMED OUT' $logfile)
	[ ! -z "$line" ] && timeoutsig=1
	line=$(grep 'Error' $logfile)
	[ ! -z "$line" ] && memoutsig=1
	
	[ $nruns -gt 1 ] && [ $timeoutsig = 1 ] && break
	[ $nruns -gt 1 ] && [ $memoutsig = 1 ] && break
done

echo -n "" > $mainresult

tailstr="DONE"
[ $memoutsig = 1 ] && tailstr="MEMOUT"
[ $timeoutsig = 1 ] && tailstr="TIMEOUT"
[ $memoutsig = 0 ] && [ $timeoutsig = 0 ] && avgSpeed=$(echo "scale=0;$avgSpeed / $nruns" | bc -l)
echo $tailstr >> $mainresult

printf "Runs                     : %-10d\n" $ntime >> $mainresult
printf "States                   : %-10d\n" $states >> $mainresult
printf "States per second (mean) : %-10d\n" $avgSpeed >> $mainresult