#!/bin/bash

export LC_NUMERIC="en_US.UTF-8"

modelpath=$1
ncores=$2
resultdir=$3
nruns=$4

[ ! -f $modelpath ] && echo "error: model '$modelpath' does not exist" && exit
[ ! -d $resultdir ] && echo "error: artifact results '$resultdir' not found" && exit

modelname=${modelpath##*/} 
modelname=${modelname%.*} 

modelbuild=generated/$modelname/$ncores
resultdest=$resultdir/spin/$modelname/$ncores

mkdir -p $modelbuild $resultdest

mainresult=$resultdest/average_all.txt

# generate SPIN binary
memlimit=310000   # practically, this value does not stop the generated binary going beyond it
logfile=$modelbuild/log_gen_and_make.txt
./spin -a $modelpath > $logfile 2>&1
mv pan* $modelbuild
gcc -o $modelbuild/$modelname  -O3 -DCOLLAPSE -DNOFAIR -DNOREDUCE -DNOBOUNDCHECK -DNCORE=$ncores -DSAFETY -DMEMLIM=$memlimit $modelbuild/pan.c > $logfile 2>&1
[ ! -f $modelbuild/$modelname ] && echo "SPIN binary for $modelname not generated" && exit
rm -rf $modelbuild/pan*

time=0
states=0
avgSpeed=0
timelimit=60     # minutes
hashtabstart=25  # this value grows up until the search depth is reached
procmemlimit=30  # GB
memoutsig=0
timeoutsig=0

[ "$modelname" = "elevator2.5" ] && nruns=1

for (( iter=0; iter<$nruns; iter++ ))
do
	ntime=$(( $iter + 1 ))
	
	printf " Running Spin on ( %-11s ) with ( %d-core ) configuration the %s-th time... \r" $modelname "$ncores" "$ntime"
	
	logfile=$resultdest/log_run${ntime}.txt

	../resourcelimit -m ${procmemlimit}000000 $modelbuild/$modelname -m50000000 -c0 -n -Q${timelimit} -w${hashtabstart} > $logfile 2>&1

	line=$(grep 'MEM CPU' $logfile)
	if [ ! -z "$line" ]; then 
		memoutsig=1
		spinstats=$( tail -n 3 $logfile )
		states=$(echo $spinstats | awk '{ print +$4 }')
		resourcestats=$( tail -n 2 $logfile )
		time=$(echo $resourcestats | awk '{ print $3 }')
		speed=$(echo "scale=0;$states / $time" | bc -l)
		avgSpeed=$speed
		break
	fi
	
	line=$(grep 'minutes exceeded' $logfile)
	[ ! -z "$line" ] && timeoutsig=1

	if [ $iter = 0 ]; then
		line=$(grep 'states, stored' $logfile)
		states=$(echo $line | awk '{ print +$1 }')
	fi
	line=$(grep 'elapsed time' $logfile)
	time=$(echo $line | awk '{ print $4 }')
	speed=$(echo "scale=0;$states / $time" | bc -l)
	avgSpeed=$(echo "$avgSpeed + $speed" | bc -l)
	
	[ $nruns -gt 1 ] && [ $timeoutsig = 1 ] && break
done

echo -n "" > $mainresult

tailstr="DONE"
[ $memoutsig = 1 ] && tailstr="MEMOUT"
[ $timeoutsig = 1 ] && tailstr="TIMEOUT"
[ $memoutsig = 0 ] && [ $timeoutsig = 0 ] && avgSpeed=$(echo "scale=0;$avgSpeed / $nruns" | bc -l)
echo $tailstr >> $mainresult

printf "Runs                     : %-10d\n" $ntime >> $mainresult
printf "States                   : %-10d\n" $states >> $mainresult
printf "States per second (mean) : %-10d\n" $avgSpeed >> $mainresult